#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Summary
-------

Test Spectrum rescaling methods

-------------------------------------------------------------------------------


"""

import astropy.units as u
import numpy as np
import pytest

import radis
from radis.misc.printer import printm
from radis.spectrum.rescale import get_recompute, get_redundant
from radis.test.utils import getTestFile
from radis.tools.database import load_spec


@pytest.mark.fast
def test_compression(verbose=True, warnings=True, *args, **kwargs):
    """Test that redundant quantities are properly inferred from already known
    spectral quantities"""

    # Get spectrum
    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    s1.conditions["thermal_equilibrium"] = True
    s1.update()

    # Analyse redundant spectral quantities
    redundant = get_redundant(s1)
    if verbose:
        print(redundant)

    assert redundant == {
        "emissivity_noslit": True,
        "radiance_noslit": True,
        "emisscoeff": True,
        "transmittance_noslit": True,
        "absorbance": True,
        "abscoeff": False,
        "xsection": True,
    }

    return True


@pytest.mark.fast
def test_update_transmittance(verbose=True, warnings=True, *args, **kwargs):
    """Test that update can correctly recompute missing quantities"""
    # TODO: add one with radiance too

    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile("N2C_specair_380nm.spec"))
    w1, T1 = s.get("transmittance_noslit")  # our reference

    if verbose:
        debug_mode = radis.config["DEBUG_MODE"]  # shows all the rescale steps taken
        radis.config["DEBUG_MODE"] = True

    # Now let's apply some update() steps

    # 1) Make sure updating doesnt change anything
    s.update()
    w2, T2 = s.get("transmittance_noslit")

    # 2) Now recompute from abscoeff
    del s._q["transmittance_noslit"]
    s.update()
    w2, T3 = s.get("transmittance_noslit")

    # 3) Now recompute from absorbance
    del s._q["transmittance_noslit"]
    del s._q["abscoeff"]
    s.update()
    w2, T4 = s.get("transmittance_noslit")

    if verbose:
        radis.config["DEBUG_MODE"] = debug_mode

    # Compare
    assert np.allclose(T1, T2)
    assert np.allclose(T1, T3)
    assert np.allclose(T1, T4)

    return True


def test_get_recompute(verbose=True, *args, **kwargs):
    """Make sure :func:`~radis.spectrum.rescale.get_recompute` works as expected

    Here, we check which quantities are needed to recompute radiance_noslit"""

    # Equilibrium
    # -----------
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    assert s.get_vars() == ["abscoeff"]
    assert s.conditions["thermal_equilibrium"]
    # At equilibrium, everything should be deduced from abscoeff
    assert set(get_recompute(s, ["radiance_noslit"])) == set(
        ("radiance_noslit", "abscoeff")
    )

    # Non Equilibrium
    # ----------------
    s.conditions["Tvib"] = 2000
    # a problem should be detected by is_at_equilibrium()
    with pytest.raises(AssertionError):
        assert not s.is_at_equilibrium(check="error")
    # force non equilibrium
    s.conditions["thermal_equilibrium"] = False

    # Now more data is needed:
    assert set(get_recompute(s, ["radiance_noslit"])) == set(
        ("abscoeff", "emisscoeff", "radiance_noslit")
    )


def test_rescale_vs_direct_computation(verbose=True, *args, **kwargs):
    """Compare spectral arrays initially computed with recomputed spectral arrays

    Notes
    -----
    Use verbose=2 to get DEBUG_MODE output
    """
    if verbose >= 2:
        DEBUG_MODE = radis.config["DEBUG_MODE"]
        radis.config["DEBUG_MODE"] = True

    # %%
    # Equilibrium Spectrum
    # -------------------
    from radis import calc_spectrum

    s = calc_spectrum(
        wavelength_min=4165,
        wavelength_max=4200,
        mole_fraction=1e-3,
        path_length=0.3,
        molecule="CO2",
        isotope="1",
        databank="HITRAN-CO2-TEST",
        Tgas=600,
        verbose=False,
    )

    s2 = s.copy()

    import numpy as np

    # 1. Compare initially computed with recomputed
    for var in ["abscoeff", "radiance_noslit"]:

        # default
        Imax_computed = s.take(var).max()
        # delete existing variables
        del s2._q[var]
        # recompute:
        Imax_recomputed = s2.take(var).max()
        # compare:
        if verbose:
            print("==>", var, Imax_computed)
            print("==>", var, Imax_recomputed)
        assert np.isclose(Imax_computed, Imax_recomputed)

    # 2. Do this systematically, for all configurations where `s` can be recomputed from.
    from radis.spectrum.rescale import _build_update_graph

    """Typical :py:func:`radis.spectrum.rescale._build_update_graph(s)` result :
    ::
        {
            "transmittance_noslit": [
                ["absorbance"],
                ["abscoeff"],
                ["absorbance"],
                ["emisscoeff"],
                ["emissivity_noslit"],
                ["transmittance"],
                ["radiance"],
                ["radiance_noslit"],
                ["xsection"],
            ],
            "absorbance": [
                ["transmittance_noslit"],
                ["abscoeff"],
                ["abscoeff"],
                ["emisscoeff"],
                ["emissivity_noslit"],
                ["transmittance"],
                ["radiance"],
                ["radiance_noslit"],
                ["transmittance_noslit"],
                ["xsection"],
            ],
            "abscoeff": [
                ["absorbance"],
                ["xsection"],
                ["absorbance"],
                ["emisscoeff"],
                ["emissivity_noslit"],
                ["transmittance"],
                ["radiance"],
                ["radiance_noslit"],
                ["transmittance_noslit"],
                ["xsection"],
            ],
            "radiance_noslit": [
                ["emisscoeff", "abscoeff"],
                ["abscoeff"],
                ["absorbance"],
                ["emisscoeff"],
                ["emissivity_noslit"],
                ["transmittance"],
                ["radiance"],
                ["transmittance_noslit"],
                ["xsection"],
            ],
            "emisscoeff": [
                ["radiance_noslit", "abscoeff"],
                ["abscoeff"],
                ["absorbance"],
                ["emissivity_noslit"],
                ["transmittance"],
                ["radiance"],
                ["radiance_noslit"],
                ["transmittance_noslit"],
                ["xsection"],
            ],
            "xsection": [["abscoeff"]],
            "emissivity_noslit": [
                ["abscoeff"],
                ["absorbance"],
                ["emisscoeff"],
                ["transmittance"],
                ["radiance"],
                ["radiance_noslit"],
                ["transmittance_noslit"],
                ["xsection"],
            ],
        }
    """
    from radis.misc.basics import all_in

    for var, vars_needed_list in _build_update_graph(s).items():
        for vars_needed in vars_needed_list:
            if all_in(vars_needed, s.get_vars()):
                if verbose:
                    print(f"Recomputing {var} from {vars_needed} only")
                s2 = s.copy()
                for v in list(s2.get_vars()):
                    if v not in vars_needed:
                        del s2._q[v]
                # recompute & compare:
                try:
                    s2.get(var)
                except NotImplementedError:
                    pass
                else:
                    assert np.isclose(s.take(var).max(), s2.take(var).max())
                    if verbose:
                        print(
                            f"Checked {var} recomputed from {vars_needed} is the same"
                        )

    #%%
    # Nonequilibrium spectrum, computed in wavenumber
    # -----------------------------------------------

    from radis import calc_spectrum

    s = calc_spectrum(
        wavenum_max=1e7 / 4165,
        wavenum_min=1e7 / 4200,
        mole_fraction=1e-3,
        path_length=0.3,
        molecule="CO2",
        isotope="1",
        databank="HITRAN-CO2-TEST",
        Tvib=2000,
        Trot=1000,
        verbose=False,
    )

    import numpy as np

    # Compare initially computed with recomputed
    for var in ["transmittance_noslit", "emisscoeff", "radiance_noslit"]:

        s2 = s.copy()

        # default
        Imax_computed = s.take(var).max()
        # delete existing variables
        del s2._q[var]
        # recompute:
        Imax_recomputed = s2.take(var).max()
        # compare:
        if verbose:
            print("==>", var, Imax_computed)
            print("==>", var, Imax_recomputed)
        assert np.isclose(Imax_computed, Imax_recomputed)

    """Typical :py:func:`radis.spectrum.rescale._build_update_graph(s)` result :
    ::

        {
            "transmittance_noslit": [["absorbance"]],
            "absorbance": [["transmittance_noslit"], ["abscoeff"]],
            "abscoeff": [["absorbance"], ["xsection"]],
            "radiance_noslit": [["emisscoeff", "abscoeff"]],
            "emisscoeff": [["radiance_noslit", "abscoeff"]],
            "xsection": [["abscoeff"]],
        }
    """

    for var, vars_needed_list in _build_update_graph(s).items():
        for vars_needed in vars_needed_list:
            if all_in(vars_needed, s.get_vars()):
                if verbose:
                    print(f"Recomputing {var} from {vars_needed} only")
                s2 = s.copy()
                for v in list(s2.get_vars()):
                    if v not in vars_needed:
                        del s2._q[v]
                # recompute & compare:
                try:
                    s2.get(var)
                except NotImplementedError:
                    pass
                else:
                    assert np.isclose(s.take(var).max(), s2.take(var).max())
                    if verbose:
                        print(
                            f"Checked {var} recomputed from {vars_needed} is the same"
                        )

    #
    if verbose >= 2:
        radis.config["DEBUG_MODE"] = DEBUG_MODE

    #%%


def test_recompute_equilibrium(verbose=True, warnings=True, plot=True, *args, **kwargs):
    """Test that spectral quantities recomputed under equilibrium assumption
    yields the same output as with non equilibrium routines when Tvib = Trot"""

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    # Get spectrum
    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s1.rescale_path_length(100)  # just for fun

    assert s1.is_at_equilibrium()
    s1.update("emisscoeff")

    # force non equilibrium calculation
    s2 = s1.copy()
    s2.conditions["thermal_equilibrium"] = False
    s2.update()
    assert (
        "emissivity_noslit" not in s2.get_vars()
    )  # just a check update() was done at nonequilibrium

    # update s1 now (at equilibrium)
    s1.update()
    assert (
        "emissivity_noslit" in s1.get_vars()
    )  # just a check update() was done at equilibrium

    s1.name = "scaled with Kirchoff law"
    s2.name = "scaled from emisscoeff + abscoeff with RTE"

    if verbose:
        print(
            "Checked that scaling at equilibrium with Kirchoff law yields the "
            + "same radiance as by solving the RTE from emisscoeff and abscoeff"
        )

    # Now Compare both update processes
    assert s1.compare_with(s2, spectra_only="radiance_noslit", plot=plot)


def test_rescale_all_quantities(verbose=True, warnings=True, *args, **kwargs):

    new_mole_fraction = 0.5
    new_path_length = 0.1

    # Get spectrum
    s0 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s0.update("all")  # start with all possible quantities in s0
    sscaled = s0.copy()
    sscaled.rescale_mole_fraction(new_mole_fraction)
    sscaled.rescale_path_length(new_path_length)
    s0.conditions["thermal_equilibrium"] = False  # to prevent rescaling with Kirchoff
    # remove emissivity_no_slit (specific to equilibrium)
    del s0._q["emissivity_noslit"]

    # Determine all quantities that can be recomputed
    if verbose >= 2:
        import radis

        DEBUG_MODE = radis.config["DEBUG_MODE"]
        radis.config["DEBUG_MODE"] = True
    from radis.spectrum.rescale import _build_update_graph, get_reachable, ordered_keys

    # ordered_keys: all spectral quantities that can be rescaled
    can_be_recomputed = get_reachable(s0)
    # can_be_recomputed: all spectra quantities that can be rescaled for this
    # particular spectrum
    update_paths = _build_update_graph(s0)
    # update_paths: which quantities are needed to recompute the others

    rescale_list = [k for k in ordered_keys if can_be_recomputed[k]]

    for quantity in rescale_list:
        all_paths = update_paths[quantity]
        if verbose:
            printm(
                "{0} can be recomputed from {1}".format(
                    quantity,
                    " or ".join(["&".join(combinations) for combinations in all_paths]),
                )
            )

        # Now let's test all paths
        for combinations in all_paths:
            if verbose:
                printm(
                    "> computing {0} from {1}".format(quantity, "&".join(combinations))
                )
            s = s0.copy()
            # Delete all other quantities
            for k in s.get_vars():
                if k not in combinations:
                    del s._q[k]

            s.update(quantity, verbose=verbose)

            # Now rescale
            s.rescale_mole_fraction(new_mole_fraction)
            s.rescale_path_length(new_path_length)

            # Compare
            assert s.compare_with(sscaled, spectra_only=quantity, plot=False)

    if verbose >= 2:
        radis.config["DEBUG_MODE"] = DEBUG_MODE


def test_xsections(*args, **kwargs):
    r"""Test that we do have

    .. math::

        exp(-\sigma \cdot N_x \cdot L) = exp(- k \cdot L)

    Therefore

    .. math::

        k = \sigma \cdot \frac{x p}{k T}

    With ``p`` total pressure and ``x`` the mole fraction.

    """
    from radis.phys.constants import k_b
    from radis.phys.units import Unit as u

    # Get spectrum
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    p = s.c["pressure"] * u("bar")
    kb = k_b * u("J/K")
    T = s.c["Tgas"] * u("K")

    N = p / kb / T

    assert s.c["pressure"] == 1.01325
    assert s.c["Tgas"] == 1500
    assert s.c["mole_fraction"] == 0.01

    assert np.isclose(N.to("1/cm3").value, 4.892e18, rtol=1e-3)

    N_x = N * 0.01

    assert np.isclose(s.take("abscoeff").max(), s.take("xsection").max() * N_x)

    # Test again for x =1

    s.rescale_mole_fraction(1)
    assert np.isclose(
        s.take("abscoeff").max().to("1/cm").value,
        (s.take("xsection").max() * N).to("1/cm").value,
    )


def test_astropy_units(verbose=True, warnings=True, *args, **kwargs):
    """This test is to assert the use of astropy units in rescale function,
    by comparing the absorbance of a spectrum rescaled with astropy units
    (in this test we use u.km) with the absorbance of original spectrum"""

    # Get precomputed spectrum
    s0 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    # Generate new spectrums by rescaling original one
    s_cm = s0.copy().rescale_path_length(100000)  # Rescale to 100000 cm = 1 km
    s_km = s0.copy().rescale_path_length(1 * u.km)  # Rescale directly to 1 astropy km

    # Get absorbance of original spectrum
    A0 = s0.get("absorbance", wunit="cm-1")

    # Get absorbance and path length of 100000cm-rescaled spectrum
    L_cm = s_cm.get_conditions()["path_length"]
    A_cm = s_cm.get("absorbance", wunit="cm-1")

    # Get absorbance and path length of 1km-rescale spectrum
    L_km = s_km.get_conditions()["path_length"]
    A_km = s_km.get("absorbance", wunit="cm-1")

    # ---------- ASSERTION ----------

    # Compare absorbances of original and 1km. They should be DIFFERENT.
    assert not np.array_equal(A0, A_km)

    # Compare absorbances of 100000cm and 1km. They should be THE SAME.
    assert np.array_equal(A_cm, A_km)

    # Compare path lengths in 1km and 100000cm. They should be THE SAME.
    assert L_cm == L_km

    if verbose:
        print(
            (
                "Astropy units work normally in the provided test case. "
                "The absorbances observed in original and rescaled spectrums "
                "follow the basis of absorption spectroscopy."
            )
        )


def _run_all_tests(verbose=True, warnings=True, *args, **kwargs):
    # test_compression(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_update_transmittance(verbose=verbose, warnings=warnings, *args, **kwargs)
    # test_get_recompute(verbose=verbose, warnings=warnings, *args, **kwargs)
    # test_rescale_vs_direct_computation(verbose=verbose, *args, **kwargs)
    # test_recompute_equilibrium(verbose=verbose, warnings=warnings, *args, **kwargs)
    # test_rescale_all_quantities(verbose=verbose, *args, **kwargs)
    # test_xsections(*args, **kwargs)
    # test_astropy_units(verbose=True, warnings=True, *args, **kwargs)
    return True


if __name__ == "__main__":
    print(("Testing test_rescale.py:", _run_all_tests(verbose=True)))



# -*- coding: utf-8 -*-
"""Tools to convert spectroscopic constants with the conventions of Hubert and
Herzberg (ω, Be, α, β, etc.) to Dunham parameters (Ykl)

References
----------

https://en.wikipedia.org/wiki/Dunham_expansion
"""

dunham2herzberg = {
    "Y01": (1, "Be"),
    "Y02": (-1, "De"),
    "Y03": (1, "He"),
    "Y04": (1, "Le"),
    "Y10": (1, "we"),
    "Y11": (-1, "alpha_e"),
    "Y12": (-1, "beta_e"),
    "Y20": (-1, "wexe"),
    "Y21": (-1, "gamma_e"),
    "Y30": (1, "weye"),
    "Y40": (1, "weze"),
}
""" dict: {Yij: (sign, coeff) }
conversion of Dunham spectroscopic coefficients to Herzberg convention """

## Invert the dictionary.
herzberg2dunham = {v: (sign, k) for k, (sign, v) in dunham2herzberg.items()}
""" dict: {Yij: (sign, coeff)
conversion of Herzberg convention to Dunham spectroscopic coefficients

See Also
--------

:py:func:`~radis.db.utils.get_dunham_coefficients`,
:py:func:`~radis.db.utils.get_herzberg_coefficients`
"""

# Name of all Herzberg coefficients
# herzberg_coefficients_elec = ['Te']
#'''list: Herzberg coefficients used to calculate pure electronic energies'''
herzberg_coefficients_vib = ["we", "wexe", "weye", "weze", "weae", "webe"]
"""list: Herzberg coefficients used to calculate pure vibrational energies"""
herzberg_coefficients_rot = ["Be", "De", "He", "Le"]
"""list: Herzberg coefficients used to calculate pure rotational energies"""
herzberg_coefficients_rovib = [
    "alpha_e",
    "beta_e",
    "gamma_e",
    "delta_e",
    "eta_e",
    "pi_e",
]
"""list: Herzberg coefficients used in vibrational-rotational coupling terms for
rovibrational energies"""

herzberg_coefficients = (
    herzberg_coefficients_vib + herzberg_coefficients_rot + herzberg_coefficients_rovib
)
"""list: Herzberg coefficients used to calculate rovibrational energies in
:func:`~radis.levels.dunham.Fv`, :func:`~radis.levels.dunham.Gv`

See Also
--------

:py:func:`~radis.db.utils.get_herzberg_coefficients`

"""


# Sanity check (test coefficients defined in conversion dictionaries are valid
# Dunham and Herzberg coefficients)
for k in dunham2herzberg:
    assert k.startswith("Y")
for k in herzberg2dunham:
    assert k in herzberg_coefficients


def get_convention(coefficients):
    """Returns if we're using the Herzberg or Dunham convention for
    spectrosopic coefficients, and returns the associated coefficients.

    Parameters
    ----------

    coefficients:
        list

    Returns
    -------

    convention: str
        'dunham' or 'herzberg'

    coefficients: dict
        list of coefficient names
    """

    from radis.db.utils import ignore_trailing_number
    from radis.misc.basics import partition

    assert len(coefficients) > 0

    herzberg_coeffs, non_herzberg_coeffs = partition(
        lambda x: ignore_trailing_number(x) in herzberg_coefficients, coefficients
    )
    for k in non_herzberg_coeffs:
        if not k.startswith("Y"):
            raise ValueError("Unexpected Spectroscopic coefficient: {0}".format(k))
    dunham_coeffs = non_herzberg_coeffs

    if len(dunham_coeffs) > 0 and len(herzberg_coeffs) > 0:
        raise ValueError(
            "Both Dunham ({0}) and Herzberg ({1}) conventions used".format(
                dunham_coeffs, herzberg_coeffs
            )
            + ". Choose one only"
        )

    if len(dunham_coeffs) > 0:
        return "dunham"
    else:
        return "herzberg"


if __name__ == "__main__":

    assert get_convention(["wexe"]) == "herzberg"
    assert get_convention(["wexe1"]) == "herzberg"
    assert get_convention(["Y01", "Y11"]) == "dunham"
    import pytest

    with pytest.raises(ValueError):
        # mixed conventions: should fail
        get_convention(["wexe", "Y01", "Y11"])



# -*- coding: utf-8 -*-
"""
Summary
-------

Models built around the :class:`~radis.spectrum.spectrum.Spectrum` class


Routine Listing
---------------

- :func:`~radis.spectrum.models.calculated_spectrum`,
- :func:`~radis.spectrum.models.experimental_spectrum`,
- :func:`~radis.spectrum.models.transmittance_spectrum`,

See Also
--------

To extract some spectral quantities from a Spectrum, and create a new Spectrum,
see the functions in :py:mod:`radis.spectrum.operations`:

- :func:`~radis.spectrum.operations.Radiance`
- :func:`~radis.spectrum.operations.Radiance_noslit`
- :func:`~radis.spectrum.operations.Transmittance`
- :func:`~radis.spectrum.operations.Transmittance_noslit`


-------------------------------------------------------------------------------


"""

import numpy as np

from radis.spectrum.spectrum import Spectrum

# %% Array-to-Spectrum functions


def calculated_spectrum(
    w,
    I,
    wunit="nm",
    Iunit="mW/cm2/sr/nm",
    conditions=None,
    cond_units=None,
    populations=None,
    name=None,
) -> Spectrum:
    """Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum`
    object that has unit conversion, plotting and slit convolution
    capabilities.

    Parameters
    ----------
    w: np.array
        wavelength, or wavenumber
    I: np.array
        intensity (no slit)
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        wavespace unit: wavelength in air (``'nm'``), wavenumber
        (``'cm-1'``), or wavelength in vacuum (``'nm_vac'``). Default ``'nm'``.
    Iunit: str
        intensity unit (can be 'counts', 'mW/cm2/sr/nm', etc...). Default
        'mW/cm2/sr/nm' (note that non-convoluted Specair spectra are in 'mW/cm2/sr/µm')


    Other Parameters
    ----------------
    conditions: dict
        (optional) calculation conditions to be stored with Spectrum. Default ``None``
    cond_units: dict
        (optional) calculation conditions units. Default ``None``
    populations: dict
        populations to be stored in Spectrum. Default ``None``
    name: str
        (optional) give a name


    Examples
    --------
    ::

        # w, I are numpy arrays for wavelength and radiance
        from radis import calculated_spectrum
        s = calculated_spectrum(w, I, wunit='nm', Iunit='W/cm2/sr/nm')     # creates 'radiance_noslit'

    .. minigallery:: radis.calculated_spectrum
        :add-heading:

    See Also
    --------

    :func:`~radis.spectrum.models.transmittance_spectrum`,
    :func:`~radis.spectrum.models.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`
    """

    return Spectrum.from_array(
        np.array(w),
        np.array(I),
        "radiance_noslit",
        wunit=wunit,
        unit=Iunit,
        conditions=conditions,
        cond_units=cond_units,
        populations=populations,
        name=name,
    )


def transmittance_spectrum(
    w, T, wunit="nm", Tunit="", conditions=None, cond_units=None, name=None
) -> Spectrum:
    """Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum`
    object that has unit conversion, plotting and slit convolution
    capabilities.

    Parameters
    ----------
    w: np.array
        wavelength, or wavenumber
    T: np.array
        transmittance (no slit)
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        wavespace unit: wavelength in air (``'nm'``), wavenumber
        (``'cm-1'``), or wavelength in vacuum (``'nm_vac'``). Default ``'nm'``.
    Iunit: str
        intensity unit. Default ``""`` (adimensioned)


    Other Parameters
    ----------------
    conditions: dict
        (optional) calculation conditions to be stored with Spectrum
    cond_units: dict
        (optional) calculation conditions units
    name: str
        (optional) give a name


    Examples
    --------
    ::

        # w, T are numpy arrays for wavelength and transmittance
        from radis import transmittance_spectrum
        s2 = transmittance_spectrum(w, T, wunit='nm')                       # creates 'transmittance_noslit'

    .. minigallery:: radis.transmittance_spectrum
        :add-heading:


    See Also
    --------

    :func:`~radis.spectrum.models.calculated_spectrum`,
    :func:`~radis.spectrum.models.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`
    """

    return Spectrum.from_array(
        np.array(w),
        np.array(T),
        "transmittance_noslit",
        wunit=wunit,
        unit=Tunit,
        conditions=conditions,
        cond_units=cond_units,
        name=name,
    )


def experimental_spectrum(
    w, I, wunit="nm", Iunit="count", conditions={}, cond_units=None, name=None
) -> Spectrum:
    """Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum`
    object that has unit conversion and plotting capabilities. Convolution is
    not available as the spectrum is assumed to have be measured experimentally
    (hence it is already convolved with the slit function)

    Parameters
    ----------
    w: np.array
        wavelength, or wavenumber
    I: np.array
        intensity
    wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
        wavespace unit: wavelength in air (``'nm'``), wavenumber
        (``'cm-1'``), or wavelength in vacuum (``'nm_vac'``). Default ``'nm'``.
    Iunit: str
        intensity unit (can be ``'count'``, 'mW/cm2/sr/nm', etc...). Default
        ``'count'`` (i.e., non calibrated output)

    Other Parameters
    ----------------
    conditions: dict
        (optional) calculation conditions to be stored with Spectrum
    cond_units: dict
        (optional) calculation conditions units
    name: str
        (optional) give a name

    Examples
    --------
    Load and plot an experimental spectrum::

        from numpy import loadtxt
        from radis import experimental_spectrum
        w, I = loadtxt('my_file.txt').T    # transpose is often useful, depending on your data.
        s = experimental_spectrum(w, I, Iunit='mW/cm2/sr/nm')             # creates 'radiance'
        s.plot()

    .. minigallery:: radis.experimental_spectrum
        :add-heading:

    See Also
    --------
    :func:`~radis.spectrum.models.calculated_spectrum`,
    :func:`~radis.spectrum.models.transmittance_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`
    """

    if np.shape(w) != np.shape(I):
        raise ValueError(
            "Wavelength {0} and intensity {1} do not have the same shape".format(
                np.shape(w), np.shape(I)
            )
        )
    return Spectrum.from_array(
        np.array(w),
        np.array(I),
        "radiance",
        wunit=wunit,
        unit=Iunit,
        conditions=conditions,
        cond_units=cond_units,
        name=name,
    )



# -*- coding: utf-8 -*-
"""
Created on Sun Aug 22 13:34:42 2020

@author: pankaj

------------------------------------------------------------------------

"""
#%%
import warnings

import pytest

import radis
from radis import SpectrumFactory, get_residual
from radis.misc.printer import printm
from radis.misc.utils import NotInstalled, not_installed_nvidia_args
from radis.misc.warning import NoGPUWarning
from radis.test.utils import getTestFile

try:
    from nvidia.cufft import __path__ as cufft_path
except ImportError:
    cufft_path = NotInstalled(*not_installed_nvidia_args)


@pytest.mark.skipif(
    isinstance(cufft_path, NotInstalled),
    reason="nvidia package not installed. Probably because on MAC OS",
)
@pytest.mark.fast
def test_eq_spectrum_emulated_gpu(
    backend="cpu-cuda", verbose=False, plot=False, *args, **kwargs
):
    """Compare Spectrum calculated in the emulated-GPU code
    :py:func:`radis.lbl.factory.SpectrumFactory.eq_spectrum_gpu` to Spectrum
    calculated with the CPU code :py:func:`radis.lbl.factory.SpectrumFactory.eq_spectrum`
    """

    print("Backend: ", backend)
    T = 1000
    p = 0.1
    wstep = 0.001
    wmin = 2284.0  # cm-1
    wmax = 2285.0  # cm-1

    # Compare GPU & CPU without sparse-LDM (not implemented in GPU yet)
    radis.config["SPARSE_WAVERANGE"] = False

    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=0.01,  # until self and air broadening is implemented
        path_length=1,  # doesnt change anything
        wstep=wstep,
        pressure=p,
        isotope="1,2",
        warnings={
            "MissingSelfBroadeningWarning": "ignore",
            "NegativeEnergiesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
            "GaussianBroadeningWarning": "ignore",
        },
    )
    sf.params.broadening_method = "fft"
    sf.load_databank(
        path=getTestFile("cdsd_hitemp_09_fragment.txt"),
        format="cdsd-hitemp",
        parfuncfmt="hapi",
    )

    s_cpu = sf.eq_spectrum(Tgas=T, name="CPU")
    s_cpu.name += f" [{s_cpu.c['calculation_time']:.2f}s]"
    s_gpu = sf.eq_spectrum_gpu(
        Tgas=T,
        backend=backend,
        name="GPU (emulate)" if backend == "cpu-cuda" else "GPU",
    )
    s_gpu.name += f"[{s_gpu.c['calculation_time']:.2f}s]"
    s_cpu.crop(wmin=2284.2, wmax=2284.8)  # remove edge lines
    s_gpu.crop(wmin=2284.2, wmax=2284.8)
    if plot:
        s_cpu.compare_with(s_gpu, spectra_only=True, plot=plot)
    assert get_residual(s_cpu, s_gpu, "abscoeff") < 1.4e-5
    assert get_residual(s_cpu, s_gpu, "radiance_noslit") < 7.3e-6
    assert get_residual(s_cpu, s_gpu, "transmittance_noslit") < 1.4e-5

    if verbose >= 2:
        print(s_gpu)
        print("\n" * 2)
        print(s_cpu)
        print("\n" * 2)
        s_gpu.print_perf_profile()
        print("\n" * 2)
        s_cpu.print_perf_profile()


@pytest.mark.skipif(
    isinstance(cufft_path, NotInstalled),
    reason="nvidia package not installed. Probably because on MAC OS",
)
@pytest.mark.needs_cuda
def test_eq_spectrum_gpu(plot=False, *args, **kwargs):
    """Compare Spectrum calculated in the GPU code
    :py:func:`radis.lbl.factory.SpectrumFactory.eq_spectrum_gpu` to Spectrum
    calculated with the CPU code :py:func:`radis.lbl.factory.SpectrumFactory.eq_spectrum`
    """
    # Ensure that GPU is not deactivated (which triggers a NoGPUWarning)
    with warnings.catch_warnings():
        warnings.simplefilter("error", category=NoGPUWarning)
        test_eq_spectrum_emulated_gpu(backend="gpu-cuda", plot=plot, *args, **kwargs)


@pytest.mark.skipif(
    isinstance(cufft_path, NotInstalled),
    reason="nvidia package not installed. Probably because on MAC OS",
)
@pytest.mark.fast
def test_multiple_gpu_calls():
    from radis import SpectrumFactory

    fixed_conditions = {
        "path_length": 1,
        "wmin": 2000,
        "wmax": 2010,
        "pressure": 0.1,
        "wstep": 0.001,
        "mole_fraction": 0.01,
    }

    sf = SpectrumFactory(**fixed_conditions, broadening_method="fft", molecule="CO")
    sf.fetch_databank("hitran")

    s1_gpu = sf.eq_spectrum_gpu(
        Tgas=300, backend="gpu-cuda", diluent={"air": 0.99}  # K  # runs on GPU
    )
    s2_gpu = sf.eq_spectrum_gpu(
        Tgas=300, backend="gpu-cuda", diluent={"air": 0.99}  # K  # runs on GPU
    )

    assert abs(s1_gpu.get_power() - s2_gpu.get_power()) / s1_gpu.get_power() < 1e-5
    assert s1_gpu.get_power() > 0


# --------------------------
if __name__ == "__main__":

    # test_eq_spectrum_gpu(plot=True)
    test_eq_spectrum_emulated_gpu(plot=True, verbose=2)

    printm("Testing GPU spectrum calculation:", pytest.main(["test_gpu.py"]))



# -*- coding: utf-8 -*-
"""

"""
import os
import shutil
from io import BytesIO
from os.path import abspath, dirname, exists, expanduser, join, split, splitext
from zipfile import ZipFile

try:
    from ..misc.config import addDatabankEntries, getDatabankEntries, getDatabankList
    from ..misc.printer import printr
    from ..misc.utils import NotInstalled, not_installed_vaex_args
    from ..misc.warning import DatabaseAlreadyExists, DeprecatedFileWarning
    from .cache_files import check_not_deprecated
    from .hdf5 import DataFileManager
except ImportError:
    if __name__ == "__main__":  # running from this file, as a script
        from radis.api.cache_files import check_not_deprecated
        from radis.api.hdf5 import DataFileManager
        from radis.misc.config import (
            addDatabankEntries,
            getDatabankEntries,
            getDatabankList,
        )
        from radis.misc.printer import printr
        from radis.misc.utils import NotInstalled, not_installed_vaex_args
        from radis.misc.warning import DatabaseAlreadyExists, DeprecatedFileWarning
    else:
        raise

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)

from datetime import date

import numpy as np
import pandas as pd
from dateutil.parser import parse as parse_date
from joblib import Parallel, delayed

try:
    from numpy.lib.npyio import DataSource
except ImportError:  # numpy <2.0.0?
    from numpy import DataSource

LAST_VALID_DATE = (
    "01 Jan 2010"  # set to a later date to force re-download of all databases
)


def get_auto_MEMORY_MAPPING_ENGINE():
    """see https://github.com/radis/radis/issues/653

    Use Vaex by default if it exists (only Python <= 3.11 as of June 2024) ,
    else use PyTables"""
    try:
        import vaex

        vaex
    except ImportError:
        return "pytables"
    else:
        return "vaex"


# Add a zip opener to the datasource _file_openers
def open_zip(zipname, mode="r", encoding=None, newline=None):
    output = BytesIO()
    with ZipFile(zipname, mode[0]) as myzip:
        fnames = myzip.namelist()
        for fname in fnames:
            output.write(myzip.read(fname))
    output.seek(0)
    return output


np.lib._datasource._file_openers._file_openers[".zip"] = open_zip


class DatabaseManager(object):
    """Line Database parser

    Parameters
    ----------
    name: str
        database name as registered in ~/radis.json
    molecule: str
    local_databases: str
        path to local database
    engine: 'vaex', 'pytables', 'h5py', or 'default'
        memory-mapping library to use with this database. If 'default' use
        the value from ~/radis.json

    Other Parameters
    ----------------
    *input for :class:`~joblib.parallel.Parallel` loading of database*
    parallel: bool
        if ``True``, use parallel loading.
        Default ``True``.
    nJobs: int
        Number of processors to use to load a database (useful for big
        databases). BE CAREFUL, no check is done on processor use prior
        to the execution ! Default ``-2``: use all but 1 processors.
        Use ``1`` for single processor.
    batch_size: int or ``'auto'``
        The number of atomic tasks to dispatch at once to each
        worker. When individual evaluations are very fast, dispatching
        calls to workers can be slower than sequential computation because
        of the overhead. Batching fast computations together can mitigate
        this. Default: ``'auto'``
    More information in :class:`joblib.parallel.Parallel`

    """

    # Should be as a close as possible to the content of the corresponding ~/radis.json entry
    # Essentially a FileManager

    def __init__(
        self,
        name,
        molecule,
        local_databases,
        engine,
        extra_params=None,
        verbose=False,
        parallel=True,
        nJobs=-2,
        batch_size="auto",
    ):
        if engine == "default":
            from radis import config

            engine = config["MEMORY_MAPPING_ENGINE"]  # 'pytables', 'vaex', 'feather'
            if engine == "auto":
                engine = get_auto_MEMORY_MAPPING_ENGINE()

        self.name = name
        self.molecule = molecule
        self.local_databases = local_databases
        self.extra_params = extra_params
        # create folder if needed
        if not exists(local_databases):
            from radis.misc.basics import make_folders

            make_folders(*split(abspath(dirname(expanduser(local_databases)))))
            make_folders(*split(abspath(expanduser(local_databases))))

        if self.is_registered():
            registered_paths = getDatabankEntries(self.name)["path"]
            for registered_path in registered_paths:
                registered_path_abspath = abspath(expanduser(registered_path)).lower()
                local_databases_abspath = abspath(expanduser(local_databases).lower())
                if not registered_path_abspath.startswith(
                    local_databases_abspath
                ):  # TODO: replace with pathlib
                    raise ValueError(
                        f"Databank `{self.name}` is already registered in radis.json but the declared path ({registered_path_abspath}) is not in the expected local databases folder ({local_databases_abspath}). Please fix/delete the radis.json entry, change the `databank_name`, or change the default local databases path entry 'DEFAULT_DOWNLOAD_PATH' in `radis.config` or ~/radis.json"
                    )

        self.downloadable = False  # by default
        self.format = ""
        self.engine = engine

        self.tempdir = join(self.local_databases, "downloads__can_be_deleted")
        self.ds = DataSource(self.tempdir)

        self.verbose = verbose

        self.parallel = parallel
        self.nJobs = nJobs
        self.batch_size = batch_size
        self.minimum_nfiles = (
            4  #: type: int. If there are less files, don't use parallel mode.
        )

    def get_filenames(self, return_reg_urls=False):
        """Get names of all files in the database (even if not downloaded yet)

        Parameters
        ----------
        return_reg_urls: (boolean)
            When the database is registered, whether to return the registered urls (``True``) or ``None`` (``False``)

        See Also
        --------
        :py:meth:`~radis.api.dbmanager.DatabaseManager.get_files_to_download`"""
        verbose = self.verbose
        local_databases = self.local_databases
        engine = self.engine

        # First; checking if the database is registered in radis.json
        if self.is_registered():
            entries = getDatabankEntries(self.name)
            try:
                assert "download_url" in entries
                assert "download_date" in entries
                assert parse_date(entries["download_date"]) > parse_date(
                    LAST_VALID_DATE
                )
            except AssertionError as err:
                raise DeprecatedFileWarning(
                    f"Database entry {self.name} not valid anymore. See above which keys are missing/wrong. Update or delete the entry in your ~/radis.json"
                ) from err
            else:
                local_files = entries["path"]
            if return_reg_urls:
                urlnames = entries["download_url"]
            else:
                urlnames = None

            # Check that local files are the one we expect :
            for f in local_files:
                local_file_abspath = abspath(expanduser(f)).lower()
                local_databases_abspath = abspath(expanduser(local_databases)).lower()
                if not local_file_abspath.startswith(local_databases_abspath):
                    raise ValueError(
                        f"Database {self.name} is inconsistent : it should be stored in {local_databases_abspath} but files registered in ~/radis.json contains {local_file_abspath}. Please fix or delete the ~/radis.json entry."
                    )

        elif self.is_downloadable():
            urlnames = self.fetch_urlnames()
            local_fnames = [
                (
                    splitext(splitext(url.split("/")[-1])[0])[
                        0
                    ]  # twice to remove .par.bz2
                    + ".h5"
                )
                for url in urlnames
            ]

            try:
                os.mkdir(local_databases)
            except OSError:
                pass
            else:
                if verbose:
                    print("Created folder :", local_databases)

            local_files = [
                abspath(
                    join(
                        local_databases,
                        self.molecule + "-" + local_fname,
                    )
                )
                for local_fname in local_fnames
            ]

            if engine == "vaex":
                local_files = [fname.replace(".h5", ".hdf5") for fname in local_files]

        else:
            raise NotImplementedError

        local_files = [expanduser(f) for f in local_files]

        return local_files, urlnames

    def fetch_urlnames(self) -> list:
        """ "This function should be overwritten by the DatabaseManager subclass

        Returns
        -------
        list: list of urlnames

        See for instance :py:class:`radis.api.hitempapi.HITEMPDatabaseManager"""

        raise NotImplementedError(
            "This function should be overwritten by the DatabaseManager subclass"
        )

    def check_deprecated_files(self, local_files, auto_remove=True):
        """Check metadata of files and remove the deprecated ones

        Unless auto_remove=False: Then raise an error"""
        verbose = self.verbose
        engine = self.engine
        for local_file in local_files:
            try:
                check_not_deprecated(
                    local_file,
                    metadata_is={},
                    metadata_keys_contain=["wavenumber_min", "wavenumber_max"],
                    engine=engine,
                )
            except DeprecatedFileWarning as err:
                if not auto_remove:
                    raise err
                else:  # delete file to regenerate it in the end of the script
                    if verbose:
                        printr(
                            "File {0} deprecated:\n{1}\nDeleting it!".format(
                                local_file, str(err)
                            )
                        )
                    os.remove(local_file)

    def get_existing_files(self, files):
        """Return files that exist among ``files``

        See Also
        --------
        :py:meth:`~radis.api.dbmanager.DatabaseManager.get_filenames`"""
        return [k for k in files if exists(k)]

    def get_missing_files(self, files):
        """Return files that do not exist among ``files``

        Note : in 'vaex' mode; if "FILE.hdf5" does not exist
        but "FILE.h5" does (a likely 'pytables' file), does
        not consider it missing so it can be converted
        automatically

        See Also
        --------
        :py:meth:`~radis.api.dbmanager.DatabaseManager.get_filenames`"""
        return [k for k in files if not exists(k)]

    def remove_local_files(self, local_files):
        # Delete existing HDF5 file
        for local_file in local_files:
            if exists(local_file):
                if self.verbose:
                    print("Removing existing file ", local_file)
                    # TODO: also clean the getDatabankList? Todo once it is in JSON format. https://github.com/radis/radis/issues/167
                os.remove(local_file)

    def is_downloadable(self):
        return self.downloadable

    def is_registered(self):
        return self.name in getDatabankList()

    def fetch_filenames(self):
        raise NotImplementedError

    def get_today(self):
        return date.today().strftime(
            "%Y-%m-%d"
        )  # only numbers avoids locale problems. Old : ("%d %b %Y")

    def register(self, dict_entries):

        # Add database to  ~/radis.json
        return register_database(
            self.name,
            dict_entries,
            self.verbose,
        )

    def get_datafile_manager(self, engine=None):
        if engine is None:
            engine = self.engine
        return DataFileManager(engine=engine)

    def download_and_parse(self, urlnames, local_files, N_files_total=None):
        if N_files_total is None:
            all_local_files, _ = self.get_filenames()
            N_files_total = len(all_local_files)

        verbose = self.verbose
        molecule = self.molecule
        parallel = self.parallel

        from time import time

        t0 = time()
        pbar_Ntot_estimate_factor = None
        if len(urlnames) != N_files_total:
            # we're only downloading a part of the database
            # expected number of lines is approximately Ntot * N_files/N_files_total
            # (this is just to give the user an expected download & parse time)
            pbar_Ntot_estimate_factor = len(urlnames) / N_files_total
        else:
            pbar_Ntot_estimate_factor = None
        Nlines_total = 0
        Ntotal_downloads = len(local_files)

        def download_and_parse_one_file(urlname, local_file, Ndownload):
            if verbose:
                inputf = urlname.split("/")[-1]
                print(
                    f"Downloading {inputf} for {molecule} ({Ndownload}/{Ntotal_downloads})."
                )

            # Check we can open the file, give the path if there is an error
            try:
                self.ds.open(urlname)
            except Exception as err:
                raise OSError(
                    f"Problem opening : {self.ds._findfile(urlname)}. See above. You may want to delete the downloaded file."
                ) from err

            # try:
            Nlines = self.parse_to_local_file(
                self.ds,
                urlname,
                local_file,
                pbar_active=(not parallel),
                pbar_t0=time() - t0,
                pbar_Ntot_estimate_factor=pbar_Ntot_estimate_factor,
                pbar_Nlines_already=Nlines_total,
                pbar_last=(Ndownload == Ntotal_downloads),
            )
            # except Exception as err:
            #     raise IOError("Problem parsing `{0}`. Check the error above. It may arise if the file wasn't properly downloaded. Try to delete it".format(self.ds._findfile(urlname))) from err

            return Nlines

        if parallel and len(local_files) > self.minimum_nfiles:
            nJobs = self.nJobs
            batch_size = self.batch_size
            if self.verbose:
                print(
                    f"Downloading and parsing {urlnames} to {local_files} "
                    + f"({len(local_files)}) files), in parallel ({nJobs} jobs)"
                )
            Nlines_total = sum(
                Parallel(n_jobs=nJobs, batch_size=batch_size, verbose=self.verbose)(
                    delayed(download_and_parse_one_file)(urlname, local_file, Ndownload)
                    for urlname, local_file, Ndownload in zip(
                        urlnames, local_files, range(1, len(local_files) + 1)
                    )
                )
            )
        else:
            for urlname, local_file, Ndownload in zip(
                urlnames, local_files, range(1, len(local_files) + 1)
            ):
                download_and_parse_one_file(urlname, local_file, Ndownload)

    def parse_to_local_file(
        self,
        opener,
        urlname,
        local_file,
        pbar_active=True,
        pbar_t0=0,
        pbar_Ntot_estimate_factor=None,
        pbar_Nlines_already=0,
        pbar_last=True,
    ) -> list:
        """This function should be overwritten by the DatabaseManager subclass

        Uncompress ``urlname`` into ``local_file``.
        Also add metadata

        Returns
        -------
        list: list of urlnames

        See for instance :py:class:`radis.api.hitempapi.HITEMPDatabaseManager"""

        raise NotImplementedError(
            "This function should be overwritten by the DatabaseManager subclass"
        )

    def clean_download_files(self):
        """Fully unzipped (and working, as it was reloaded): clean files

        Note : we do not let np.DataSource clean its own tempfiles; as
        they may be downloaded but not fully parsed / registered."""
        if exists(self.tempdir):
            try:
                shutil.rmtree(self.tempdir)
            except PermissionError as err:
                if self.verbose >= 3:
                    from radis.misc.printer import printr

                    printr(
                        f"... couldnt delete downloaded cache files {self.tempdir}: {str(err)}"
                    )
            else:
                if self.verbose >= 3:
                    from radis.misc.printer import printg

                    printg(f"... removed downloaded cache files for {self.tempdir}")

    def load(
        self,
        local_files,
        columns=None,
        lower_bound=[],
        upper_bound=[],
        within=[],
        output="pandas",
    ):
        """
        Other Parameters
        ----------------
        columns: list of str
            list of columns to load. If ``None``, returns all columns in the file.
        output: 'pandas', 'vaex', 'jax'
            format of the output DataFrame. If ``'jax'``, returns a dictionary of
            jax arrays.
        lower_bound: list of tuples [(column, lower_bound), etc.]
            ::

                lower_bound =[("wav", load_wavenum_min)]
        upper_bound_bound: list of tuples [(column, upper_bound), etc.]
            ::

                upper_bound=[("wav", load_wavenum_max)]
        within: list of tuples [(column, within_list), etc.]
            ::

                within=[("iso", isotope.split(","))]
        """
        engine = self.engine
        mgr = self.get_datafile_manager()
        if engine in ["pytables", "feather"]:
            df_all = []
            for local_file in local_files:
                df_all.append(
                    mgr.load(
                        local_file,
                        columns=columns,
                        lower_bound=lower_bound,
                        upper_bound=upper_bound,
                        within=within,
                        output=output,
                    )
                )
            return pd.concat(df_all)

        elif engine == "vaex":
            # vaex can open several files at the same time:
            return mgr.load(
                local_files,
                columns=columns,
                lower_bound=lower_bound,
                upper_bound=upper_bound,
                within=within,
                output=output,
            )
        else:
            raise NotImplementedError(engine)

    def plot(self, local_files, isotope, wavenum_min, wavenum_max):
        """Convenience function to plot linestrengths of the database"""

        df = self.load(
            local_files, isotope, wavenum_min, wavenum_max, columns=["wav", "int"]
        )
        df.plot("wav", "int")

    def get_nrows(self, local_file):
        """Get number of rows (without loading all DataFrame)"""
        engine = self.engine
        local_file = expanduser(local_file)
        if engine == "pytables":
            with pd.HDFStore(local_file, "r") as store:
                nrows = store.get_storer("df").nrows

        elif engine == "vaex":
            # by default vaex does not load everything
            df = vaex.open(local_file)
            nrows = len(df)
            df.close()

        elif engine in ["h5py"]:
            raise NotImplementedError
        else:
            raise ValueError(engine)
        return nrows

    def get_columns(self, local_file):
        """Get all columns using DataFileManager class get_columns function"""
        return self.get_datafile_manager().get_columns(local_file)

    def add_column(self, df, key, value):
        """Create column ``key`` in DataFrame or dictionary ``df`` with value ``value``"""
        from radis.misc.basics import is_number

        if is_number(self.alpha_ref):
            if isinstance(df, vaex.dataframe.DataFrameLocal):
                # see https://github.com/vaexio/vaex/pull/1570
                df[key] = vaex.vconstant(float(value), length=df.length_unfiltered())
            else:
                df[key] = value
        else:
            df[key] = value

    def rename_columns(self, df, rename_dict):
        """Example::

        mdb.rename_columns(df, {"nu_lines":"wav"})
        """
        if not isinstance(vaex, NotInstalled) and isinstance(
            df, vaex.dataframe.DataFrameLocal
        ):
            for k, v in rename_dict.items():
                df.rename(k, v)
        elif isinstance(df, pd.DataFrame):
            df.rename(columns=rename_dict, inplace=True)
        elif isinstance(df, dict):
            for k, v in rename_dict.items():
                df[v] = df.pop(k)


def register_database(databank_name, dict_entries, verbose):
    """Add to registered databases in RADIS config file.

    If database exists, assert it has the same entries.

    Parameters
    ----------
    databank_name: str
        name of the database in :ref:`~/radis.json config file <label_lbl_config_file>`

    Other Parameters
    ----------------
    verbose: bool

    Returns
    -------
    None:
        adds to :ref:`~/radis.json <label_lbl_config_file>` with all the input
        parameters. Also adds ::

            format : "hitemp-radisdb"
            parfuncfmt : "hapi"   # TIPS-2017 for equilibrium partition functions

        And if the molecule is in :py:attr:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`::

            levelsfmt : "radis"   # use RADIS spectroscopic constants for rovibrational energies (nonequilibrium)

    """

    # Register database in ~/radis.json to be able to use it with load_databank()
    try:
        addDatabankEntries(databank_name, dict_entries)
    except DatabaseAlreadyExists as err:
        # Check that the existing database had the same entries
        try:
            from radis.misc.config import getDatabankEntries

            for k, v in getDatabankEntries(databank_name).items():
                if k == "download_date":
                    continue
                assert dict_entries[k] == v
            # TODO @dev : replace once we have configfile as JSON (https://github.com/radis/radis/issues/167 )
        except AssertionError:
            raise DatabaseAlreadyExists(
                f"{databank_name} already exists in your ~/radis.json config file, "
                + f"with different key `{k}` : `{v}` (~/radis.json) ≠ `{dict_entries[k]}` (new). "
                + "If you're sure of what you're doing, fix the registered database in ~/radis.json. "
                + "Else, remove it from your config file, or choose a different name "
                + "for the downloaded database with `fetch_hitemp(databank_name=...)`, "
                + "and restart."
            ) from err
        else:  # no other error raised
            if verbose:
                print(
                    f"{databank_name} already registered in ~/radis.json config file, with the same parameters."
                )



# -*- coding: utf-8 -*-
"""
Functions to read/write Spectrum objects under HDF5 format
"""

from radis.io.hdf5 import DataFileManager

# Convert


def spec2hdf(s, file, engine="pytables"):
    """Convert Spectrum s to HDF5 file"""

    df_arrays = s.to_pandas(copy=False)

    # Store
    mgr = DataFileManager(engine=engine)

    # Store spectral arrays and lines as Datasets

    mgr.write(file, df_arrays, key="arrays", append=False, data_columns=["wavespace"])
    mgr.add_metadata(file, s.units, key="arrays")

    if s.lines:
        mgr.write(file, s.lines, key="lines", append=True)

    # Store the rest as metadata

    if s.populations:
        mgr.add_metadata(
            file, s.populations, key="populations", create_empty_dataset=True
        )

    mgr.add_metadata(file, s.conditions, key="conditions", create_empty_dataset=True)

    if s.references:
        mgr.add_metadata(
            file, s.references, key="references", create_empty_dataset=True
        )


def hdf2spec(file, wmin=None, wmax=None, wunit=None, columns=None, engine="pytables"):
    """Read HDF5 file into a Spectrum object"""
    from radis import Spectrum

    mgr = DataFileManager(engine=engine)

    # Range selection
    # TODO : convert if stored in wavelength/wavenumber
    if wmin is not None or wmax is not None:
        if engine == "pytables":
            where = []
            if wmin is not None:
                where.append(f"wavespace > {wmin}")
            if wmax is not None:
                where.append(f"wavespace < {wmax}")
        elif engine == "vaex":
            #  Selection is done after opening the file in vaex
            where = None
        else:
            raise NotImplementedError(engine)
    else:
        where = None

    # Column selection
    # ... always load 'wavespace'
    if columns and "wavespace" not in columns:
        columns.append("wavespace")

    df2 = mgr.load(file, columns=columns, where=where, key="arrays")

    # Convert to Pandas df (if vaex)
    if engine == "vaex":
        if wmin is not None and wmax is not None:
            df2.select((df2.wavespace > wmin) & (df2.wavespace < wmax))
            selection = True
        elif wmin is not None:
            df2.select(df2.wavespace > wmin)
            selection = True
        elif wmax is not None:
            df2.select(df2.wavespace < wmax)
            selection = True
        else:
            selection = False
        df2 = df2.to_pandas_df(selection=selection, column_names=columns)

    units2 = mgr.read_metadata(file, key="arrays")
    conditions2 = mgr.read_metadata(file, key="conditions")
    try:
        populations2 = mgr.read_metadata(file, key="populations")
    except KeyError:
        populations2 = None
    try:
        references2 = mgr.read_metadata(file, key="references")
    except KeyError:
        references2 = None

    try:
        lines2 = mgr.load(file, key="lines")
    except (KeyError, OSError):  # KeyError in PyTables ; OSError in Vaex
        lines2 = None

    s2 = Spectrum(
        {k: v.values for k, v in dict(df2).items()},
        lines=lines2,
        units=units2,
        conditions=conditions2,
        references=references2,
        populations=populations2,
    )

    return s2



"""Molecular database (MDB) class

   * MdbExomol is the MDB for ExoMol

Initial code borrowed from the `Exojax <https://github.com/HajimeKawahara/exojax>`__
code (which you should also have a look at!), by @HajimeKawahara, under MIT License.

"""

import os
import pathlib
import warnings

import numpy as np

from radis.api.dbmanager import DatabaseManager, get_auto_MEMORY_MAPPING_ENGINE
from radis.db.classes import EXOMOL_MOLECULES, EXOMOL_ONLY_ISOTOPES_NAMES

EXOMOL_URL = "http://www.exomol.com/db/"

import bz2
import re
from urllib.request import HTTPError, urlopen

import pandas as pd
from bs4 import BeautifulSoup

from radis.api.hdf5 import vaexsafe_colname


def exact_molname_exomol_to_simple_molname(exact_exomol_molecule_name):
    """convert the exact molname (used in ExoMol) to the simple molname.

    Args:
        exact_exomol_molecule_name: the exact exomol molecule name

    Returns:
        simple molname

    Examples:
        >>> print(exact_molname_exomol_to_simple_molname("12C-1H4"))
        >>> CH4
        >>> print(exact_molname_exomol_to_simple_molname("23Na-16O-1H"))
        >>> NaOH
        >>> print(exact_molname_exomol_to_simple_molname("HeH_p"))
        >>> HeH_p
        >>> print(exact_molname_exomol_to_simple_molname("trans-31P2-1H-2H")) #not working
        >>> Warning: Exact molname  trans-31P2-1H-2H cannot be converted to simple molname
        >>> trans-31P2-1H-2H
        >>> print(exact_molname_exomol_to_simple_molname("1H-2H-16O"))
        >>> H2O
    """

    try:
        molname_simple = _molname_simple_no_exception(exact_exomol_molecule_name)
        return molname_simple
    except:
        print(
            "Warning: Exact molname ",
            exact_exomol_molecule_name,
            "cannot be converted to simple molname",
        )
        return exact_exomol_molecule_name


def _molname_simple_no_exception(exact_exomol_molecule_name):
    t = exact_exomol_molecule_name.split("-")
    molname_simple = ""
    for ele in t:
        alp = "".join(re.findall(r"\D", ele))
        num0 = re.split("[A-Z]", ele)[1]
        if num0.isdigit():
            num = num0
        else:
            num = ""
        molname_simple = molname_simple + alp + num

    # ExoJAX Issue #528
    if molname_simple == "HHO":
        molname_simple = "H2O"
    return molname_simple


def read_def(deff):
    """Exomol IO for a definition file

    Parameters
    ----------
    deff: definition file

    Returns
    -------
    n_Texp : float
        temperature exponent
    alpha_ref : float
        broadening parameter
    molmass : float
        molecular mass
    numinf : List[float]
        limit points (``[w(0), w(1), ..., w(n)]``, n+1 elements) defining
        the spectral ranges appearing in the name of *.trans.bz2 files
        (``["w(0)-w(1)", "w(1)-w(2)", ..., w(n-1)-w(n)]``, n elements)
    numtag : List[str]
        tag for wavelength ranges.

    Note:
        For some molecules, ExoMol provides multiple trans files. numinf and numtag are the ranges and identifiers for the multiple trans files.

    """

    dat = pd.read_csv(deff, sep="#", names=("VAL", "COMMENT"))
    alpha_ref = None
    # texp = None
    n_Texp = None
    ntransf = 1
    maxnu = 0.0
    quantum_labels = []
    unc = False
    for i, com in enumerate(dat["COMMENT"]):
        if "Default value of Lorentzian half-width" in com:
            alpha_ref = float(dat["VAL"][i])
        elif "Default value of temperature exponent" in com:
            n_Texp = float(dat["VAL"][i])
        elif "No. of transition files" in com:
            ntransf = int(dat["VAL"][i])
        elif "Maximum wavenumber (in cm-1)" in com:
            maxnu = float(dat["VAL"][i])
            # maxnu=20000.0
        elif "Isotopologue mass (Da) and (kg)" in com:
            molmass = float(dat["VAL"][i].split()[0])  # in Da (atomic unit)

        elif "Lifetime availability" in com:
            lifetime = int(dat["VAL"][i]) == 1
        elif "Lande g-factor availability" in com:
            lande = int(dat["VAL"][i]) == 1
        elif "Quantum label" in com:
            quantum_labels.append(dat["VAL"][i].strip(" "))
        elif "Uncertainty availability" in com:
            unc = int(dat["VAL"][i]) == 1

    # SOME DEF FILES CONTAINS ERRORS. THESE ARE THE EXCEPTIONS
    if deff.stem == "1H-35Cl__HITRAN-HCl":
        quantum_labels = ["v"]
        # See https://github.com/HajimeKawahara/exojax/issues/330
    if deff.stem == "16O-1H__MoLLIST":
        quantum_labels = ["e/f", "v", "F1/F2", "Es"]

    if deff.stem == "1H-2H-16O__VTT":
        numinf = wavenumber_range_HDO_VTT()
    else:
        numinf = compute_wavenumber_ranges(ntransf, maxnu)

    numtag = wavenumber_tag(numinf)

    output = {
        "n_Texp": n_Texp,
        "alpha_ref": alpha_ref,
        "molmass": molmass,
        "numinf": numinf,
        "numtag": numtag,
        "quantum_labels": quantum_labels,  # array
        "Landé": lande,  # bool
        "lifetime": lifetime,  # bool
        "unc": unc,  # bool uncertainty of line center availability
    }
    return output


def wavenumber_range_HDO_VTT():
    """wave number range for HDO VTT as an exception

    Returns:
        float: numinf
    """
    numinf = np.array(
        [
            0.0,
            250.0,
            500.0,
            750.0,
            1000.0,
            1500.0,
            2000.0,
            2250.0,
            2750.0,
            3500.0,
            4500.0,
            5500.0,
            7000.0,
            9000.0,
            14000.0,
            20000.0,
            26000.0,
        ]
    )
    return numinf


def compute_wavenumber_ranges(ntransf, maxnu):
    """wavenumber range for general case

    Args:
        ntransf: number of the trans files
        maxnu: maximum wavenumber

    Returns:
        float: numinf
    """
    if ntransf > 1:
        dnufile = maxnu / ntransf
        numinf = dnufile * np.array(range(ntransf + 1))
        wavenumber_tag(numinf)
    else:
        numinf = None
    return numinf


def wavenumber_tag(numinf):
    """convert numinf to numtag (used in the name of trans file)


    Args:
        numinf (_type_): wavenumber values for the range

    Returns:
        float: numtag wavenumber tag (such as 00000-00500)
    """
    if numinf is None:
        return ""
    numtag = []
    for i in range(len(numinf) - 1):
        imin = "{:05}".format(int(numinf[i]))
        imax = "{:05}".format(int(numinf[i + 1]))
        numtag.append(imin + "-" + imax)
    return numtag


def read_pf(pff):
    """Exomol IO for partition file

    Note:
        T=temperature
        QT=partition function

    Args:
        pff: partition file
    Returns:
        partition data in pandas DataFrame

    """
    dat = pd.read_csv(pff, sep=r"\s+", names=("T", "QT"))
    return dat


def read_trans(transf, engine="vaex"):
    """Exomol IO for a transition file

    Notes
    -----

    Transf format ::

        i_upper=Upper state counting number
        i_lower=Lower state counting number
        A=Einstein coefficient in s-1
        nu_lines=transition wavenumber in cm-1

    See Table 12 in https://arxiv.org/pdf/1603.05890.pdf [Exomol-2016]_

    Parameters
    ----------
    transf: transition file
    engine: parsing engine to use ('vaex', 'csv')

    Returns
    -------
    transition data in vaex/pandas DataFrame

    """
    if engine == "vaex":
        import vaex

        try:  # bz2 compression
            dat = vaex.from_csv(
                transf,
                compression="bz2",
                sep=r"\s+",
                names=("i_upper", "i_lower", "A", "nu_lines"),
                convert=False,  #  file is created by MdbMol
            )
        except:
            try:
                dat = vaex.read_csv(
                    transf,
                    sep=r"\s+",
                    names=("i_upper", "i_lower", "A", "nu_lines"),
                    convert=False,  #  file is created by MdbMol
                )
            except Exception as err:
                raise Exception(
                    f"Error reading {transf}. Either the file does not exist on exomol.com, or was corrupted? You can try to delete it."
                ) from err
    elif engine == "csv":
        try:  # bz2 compression
            dat = pd.read_csv(
                transf,
                compression="bz2",
                sep=r"\s+",
                names=("i_upper", "i_lower", "A", "nu_lines"),
            )
        except:
            try:
                dat = pd.read_csv(
                    transf, sep=r"\s+", names=("i_upper", "i_lower", "A", "nu_lines")
                )
            except Exception as err:
                raise Exception(
                    f"Error reading {transf}. Either the file does not exist on exomol.com, or was corrupted? You can try to delete it."
                ) from err
    else:
        raise NotImplementedError(engine)

    return dat


def read_states(
    statesf, dic_def, engine="vaex", skip_optional_data=True, print_states=False
):
    """Exomol IO for a state file

    Notes
    -----

    States f format ::

        i=state counting number
        E=state energy
        g=state degeneracy
        J=total angular momentum

    See Table 11 in https://arxiv.org/pdf/1603.05890.pdf

    Parameters
    ----------
    statesf: str
        state file
    dic_def: dict
        Info from def file to read extra quantum numbers
    engine: str
        parsing engine to use ('vaex', 'csv')
    skip_optional_data: bool
        If False, fetch all fields which are marked as available in the ExoMol definition
        file. If True, load only the first 4 columns of the states file
        ("i", "E", "g", "J"). The structure of the columns above 5 depend on the
        the definitions file (*.def) and the Exomol version.
        If ``skip_optional_data=False``, two errors may occur:

            - a field is marked as present/absent in the *.def field but is
              absent/present in the *.states file (ie both files are inconsistent).
            - in the updated version of Exomol, new fields have been added in the
              states file of some species. But it has not been done for all species,
              so both structures exist. For instance, the states file of
              https://exomol.com/data/molecules/HCl/1H-35Cl/HITRAN-HCl/ follows the
              structure described in [1]_, unlike the states file of
              https://exomol.com/data/molecules/NO/14N-16O/XABC/ which follows the
              structure described in [2]_.
    print_states: bool
        print some info. When .def file looks inconsistent with .states file, turn ON and report them in Issue.


    Returns
    -------
    states data in pandas DataFrame

    If ``'vaex'``, also writes a local hdf5 file `statesf.with_suffix('.hdf5')`


    References
    ----------

    .. [1] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Barton, E. J., Chubb, K. L., Coles, P. A., … Zak, E. (2016). The ExoMol database: molecular line lists for exoplanet and other hot atmospheres. https://doi.org/10.1016/j.jms.2016.05.002
    .. [2] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Clark, V. H. J., Chubb, K. L., Conway, E. K., … Yurchenko, O. P. (2020). The 2020 release of the ExoMol database: Molecular line lists for exoplanet and other hot atmospheres. Journal of Quantitative Spectroscopy and Radiative Transfer, 255, 107228. https://doi.org/10.1016/j.jqsrt.2020.107228

    """
    # we read first 4 columns for ("i", "E", "g", "J"),
    # skip lifetime, skip Landé g-factor,
    # read quantum numbers
    N_otherfields = np.size(dic_def["quantum_labels"])
    quantum_labels = dic_def["quantum_labels"]

    mandatory_fields = ("i", "E", "g", "J")
    N_mandatory_fields = len(mandatory_fields)
    mandatory_usecol = np.arange(N_mandatory_fields)
    if skip_optional_data:
        usecol = mandatory_usecol
        names = mandatory_fields
    else:
        label = np.array(["unc", "lifetime", "Lande"])
        mask = np.array([dic_def["unc"], dic_def["Landé"], dic_def["lifetime"]])
        N_except = np.sum(mask)
        usecol = np.arange(0, N_mandatory_fields + N_except + N_otherfields)
        names = mandatory_fields + tuple(label[mask]) + tuple(quantum_labels)
        # The definitions file (*.def) specifies which fields are available
        # in the states fiel (*.states). Check the number of available fields
        # (see *.def) match the numbers of columns in the states file (*.states).
        # Otherwise, both files are inconsistent.
        with bz2.open(statesf, "rb") as f:
            firstline = f.readline().decode("utf-8")  # read 1dt line
            splitline = [x for x in re.split(r"\s+", firstline) if x != ""]
            # splitline = re.split(r"\s+", firstline)
            f.close()
        if len(splitline) != len(names):
            warnings.warn(
                "There appear to be further additional columns in .state file,"
                + "which are not defined in .def file. We skip them."
            )

    if engine == "vaex":
        import vaex

        # TODO Refactor: move in DataFileManager
        try:
            dat = vaex.from_csv(
                statesf,
                compression="bz2",
                sep=r"\s+",
                usecols=usecol,
                names=names,
                convert=False,  # written in MolDB
            )
        except:
            try:
                dat = vaex.read_csv(
                    statesf,
                    sep=r"\s+",
                    usecols=usecol,
                    names=names,
                    convert=False,  # written in MolDB
                )

            except Exception as err:
                raise Exception(
                    f"Error reading {statesf}. Either the file does not exist on exomol.com, or was corrupted? You can try to delete it."
                ) from err
    elif engine == "csv":
        try:
            dat = pd.read_csv(
                statesf, compression="bz2", sep=r"\s+", usecols=usecol, names=names
            )
        except:  #!!!TODO What was the expected error?
            dat = pd.read_csv(statesf, sep=r"\s+", usecols=usecol, names=names)
    else:
        raise NotImplementedError(engine)
    if print_states:
        print(dat)
    return dat


def pickup_gE(states, trans, dic_def, skip_optional_data=True, engine="vaex"):
    """extract g_upper (gup), E_lower (elower), and J_lower and J_upper from states
    DataFrame and insert them into the transition DataFrame.

    Parameters
    ----------
    states: states DataFrame  - the i, E, g, J are in the 4 first columns
    trans: transition numpy array
    dic_def: Informations about additional quantum labels
    skip_optional_data: bool . If True fetch all quantum labels in dic_def['quantum_labels'] from states into transitions (_l for lower, _u for upper states)

    Returns
    -------
    A, nu_lines, elower, gup, jlower, jupper, mask, **quantum_labels

    Notes
    -----
    We first convert pandas DataFrame to ndarray. The state counting numbers in states DataFrame is used as indices of the new array for states (newstates). We remove the state count numbers as the column of newstate, i.e. newstates[:,k] k=0: E, 1: g, 2: J. Then, we can directly use the state counting numbers as mask.

    States by default has columns ::

        #       i      E             g    J    v
        0       1      0.0           1    0    0
        1       2      1.448467      3    1    0
        2       3      4.345384      5    2    0
        3       4      8.690712      7    3    0


    """

    ### Step 1. Essential quantum number for spectra
    # ----------------------------------------------

    def map_add(trans, col, new_col, trans_key, states_key="i"):
        """Lookup `key` in states and add it in trans, using the level ``i``
        for upper and lower state

        Examples
        --------
        ::

            trans = map_add(trans, "E", "E_lower", "i_lower")
        """
        if engine == "pytables":
            # Rename the columns in the states DataFrame
            states_map = states.copy()
            states_map.rename(
                columns={col: new_col, states_key: trans_key}, inplace=True
            )
            states_map = states_map[[new_col, trans_key]]  # drop useless columns
            trans = trans.join(states_map.set_index(trans_key), on=trans_key)
        elif engine == "vaex":
            col = vaexsafe_colname(col)
            new_col = vaexsafe_colname(new_col)

            # WIP. First implementation with join(). Use Map() will probably be faster.
            trans.join(
                states[states_key, col],
                left_on=trans_key,
                right_on=states_key,
                inplace=True,
            )
            trans.drop(states_key, inplace=True)
            trans.rename(col, new_col)
        return trans

    trans = map_add(trans, "g", "gup", "i_upper")
    trans = map_add(trans, "J", "jlower", "i_lower")
    trans = map_add(trans, "J", "jupper", "i_upper")
    trans = map_add(trans, "E", "elower", "i_lower")

    def has_nan(column):
        try:  # Vaex
            return column.countnan() > 0
        except AttributeError:  # Pandas
            return column.hasnans

    if not "nu_lines" in trans or has_nan(trans["nu_lines"]):
        trans = map_add(trans, "E", "eupper", "i_upper")
        trans["nu_lines"] = trans["eupper"] - trans["elower"]

    ### Step 2. Extra quantum numbers (e/f parity, vib and rot numbers)
    # -----------------------------------------------------------------
    if not skip_optional_data:
        for q in dic_def["quantum_labels"]:
            trans = map_add(trans, q, f"{q}_l", "i_lower")
            trans = map_add(trans, q, f"{q}_u", "i_upper")

    return trans


# def pickup_gEslow(states, trans):
#     """Slow version to extract g_upper (gup) and E_lower (elower) from states DataFrame and insert them to transition DataFrame.

#     Note:
#        This code is the (thousands times) slower version of pickup_gE. However, we did not remove this one just because this version is much easier to understand.

#     Args:
#        states: states pandas DataFrame
#        trans: transition pandas DataFrame

#     Returns:
#        A, nu_lines, elower, gup

#     """
#     import tqdm

#     E = states["E"].values
#     g = states["g"].values

#     # insert new columns in transition array
#     trans["gup"] = 0
#     trans["elower"] = 0.0

#     for k, i in tqdm.tqdm(enumerate(states["i"])):
#         # transition upper state
#         mask_upper = trans["i_upper"] == i
#         trans["gup"][mask_upper] = g[k]
#         # transition lower state
#         mask_lower = trans["i_lower"] == i
#         trans["elower"][mask_lower] = E[k]

#     A = trans["A"].to_numpy()
#     nu_lines = trans["nu_lines"].to_numpy()
#     elower = trans["elower"].to_numpy()
#     gup = trans["gup"].to_numpy()
#     return A, nu_lines, elower, gup


def read_broad(broadf):
    """Reading broadening file (.broad)
    Parameters
    ----------
    broadf: .broad file

    Returns
    -------
    broadening info in bdat form (pandas), defined by this instance.

    Notes
    -----
    See Table 16 in https://arxiv.org/pdf/1603.05890.pdf
    """
    bdat = pd.read_csv(
        broadf,
        sep=r"\s+",
        names=(
            "code",
            "alpha_ref",
            "n_Texp",
            "jlower",
            "jupper",
            "kalower",
            "kclower",
            "kaupper",
            "kcupper",
            "v1lower",
            "v2lower",
            "v3lower",
            "v1upper",
            "v2upper",
            "v3upper",
        ),
    )

    return bdat


def check_code_level(bdat):
    """checking code level in .broad
    Args:
        bdat: exomol .broad data given by exomolapi.read_broad

    Returns:
        code level: None, a0, a1, other codes unavailable currently,
        if a0 and a1 are available, a1 is returned.
        the other cases returns None
    """
    input_array = np.unique(np.array(bdat["code"]))
    if np.array_equal(input_array, np.array(["a0"])):
        return "a0"
    elif np.array_equal(input_array, np.array(["a1"])):
        return "a1"
    elif np.array_equal(np.sort(input_array), np.array(["a0", "a1"])):
        return "a1"
    return None


def make_j2b(bdat, alpha_ref_default=0.07, n_Texp_default=0.5, jlower_max=None):
    """compute j2b (code a0, map from jlower to alpha_ref)

    Args:
        bdat: exomol .broad data given by exomolapi.read_broad
        alpha_ref_default: default value
        n_Texp_default: default value
        jlower_max: maximum number of jlower
    Returns:
        j2alpha_ref[jlower] provides alpha_ref for jlower
        j2n_Texp[jlower]  provides nT_exp for jlower
    """
    # a0
    cmask = bdat["code"] == "a0"
    jlower_arr = np.array(bdat["jlower"][cmask], dtype=int)
    alpha_ref_arr = np.array(bdat["alpha_ref"][cmask])
    n_Texp_arr = np.array(bdat["n_Texp"][cmask])

    # Determine the array size based on jlower_max
    if jlower_max is None:
        Nblower = np.max(jlower_arr) + 1
    else:
        Nblower = np.max([jlower_max, np.max(jlower_arr)]) + 1

    # Initialize arrays with default alpha_ref and n_Texp
    j2alpha_ref = np.full(Nblower, alpha_ref_default)
    j2n_Texp = np.full(Nblower, n_Texp_default)

    # Populate the mapping arrays using known broadening coefficients
    j2alpha_ref[jlower_arr] = alpha_ref_arr
    j2n_Texp[jlower_arr] = n_Texp_arr

    # Raise a minor warning if default values are used for high J values
    if Nblower > (np.max(jlower_arr) + 1):
        import warnings

        from radis.misc.warning import AccuracyWarning

        warnings.warn(
            AccuracyWarning(
                "The default broadening parameter (alpha = {2} cm^-1 and n = {3}) are used for J'' > {1} up to J'' = {0}".format(
                    Nblower, np.max(jlower_arr), alpha_ref_default, n_Texp_default
                )
            )
        )

    return j2alpha_ref, j2n_Texp


def make_jj2b(bdat, j2alpha_ref_def, j2n_Texp_def, jupper_max=None):
    """compute jj2b (code a1, map from (jlower, jupper) to alpha_ref and n_Texp)

    Args:
        bdat: exomol .broad data given by exomolapi.read_broad
        j2alpha_ref_def: default value from a0
        j2n_Texp_def: default value from a0
        jupper_max: maximum number of jupper
    Returns:
        jj2alpha_ref[jlower,jupper] provides alpha_ref for (jlower, jupper)
        jj2n_Texp[jlower,jupper]  provides nT_exp for (jlower, jupper)
    Note:
        The pair of (jlower, jupper) for which broadening parameters are not given, jj2XXX contains None.
    """
    # a1
    cmask = bdat["code"] == "a1"
    jlower_arr = np.array(bdat["jlower"][cmask], dtype=int)
    jupper_arr = np.array(bdat["jupper"][cmask], dtype=int)
    alpha_ref_arr = np.array(bdat["alpha_ref"][cmask])
    n_Texp_arr = np.array(bdat["n_Texp"][cmask])

    if jupper_max is None:
        Nbupper = np.max(jupper_arr) + 1
    else:
        Nbupper = np.max([jupper_max, np.max(jupper_arr)]) + 1

    jj2alpha_ref = j2alpha_ref_def[:, np.newaxis] * np.ones(Nbupper)
    jj2n_Texp = j2n_Texp_def[:, np.newaxis] * np.ones(Nbupper)

    jj2alpha_ref[jlower_arr, jupper_arr] = alpha_ref_arr
    jj2n_Texp[jlower_arr, jupper_arr] = n_Texp_arr

    return jj2alpha_ref, jj2n_Texp


def get_exomol_full_isotope_name(molecule, isotope):
    """Get full isotope name for ``molecule`` and ``isotope`` number

    Parameters
    ----------
    molecule: str
    isotope: int
        terrestrial abundance

    Examples
    --------
    ::

        get_exomol_full_isotope_name("CH4", 1)
        >>> '12C-1H4'

    See Also
    --------
    :py:func:`~radis.api.exomolapi.get_exomol_database_list`"""

    if molecule not in EXOMOL_MOLECULES:
        raise ValueError(
            f"Molecule {molecule} not available in ExoMol. Change database, or choose one of ExoMol available molecules: {sorted(EXOMOL_MOLECULES)}"
        )

    if (molecule, isotope) in EXOMOL_ONLY_ISOTOPES_NAMES:
        return EXOMOL_ONLY_ISOTOPES_NAMES[(molecule, isotope)]
    else:
        # Read and convert from HITRAN molecules
        from radis.db.molparam import MolParams

        mp = MolParams()
        return mp.get(molecule, isotope, "isotope_name_exomol")


def get_list_of_known_isotopes(molecule):
    """find all isotopes until error ensues"""

    i = 1
    isotope_list = []
    while True:
        try:
            iso_name = get_exomol_full_isotope_name(molecule, i)
        except:
            break
        else:
            isotope_list.append(iso_name)
        finally:
            i += 1
    return isotope_list


def get_exomol_database_list(molecule, isotope_full_name=None):
    """Parse ExoMol website and return list of available databases, and recommended database

    Parameters
    ----------
    molecule: str
    isotope_full_name: str
        isotope full name (ex. ``12C-1H4`` for CH4,1). Get it from
        :py:func:`radis.api.exomolapi.get_exomol_full_isotope_name`

    Returns
    -------
    list of databases, database recommended by ExoMol

    Examples
    --------
    Get CH4 from ExoMol :
    ::

        databases, recommended = get_exomol_database_list("CH4", "12C-1H4")
        >>> ['xsec-YT10to10', 'YT10to10', 'YT34to10'], 'YT34to10'

    Or combine with :py:func:`~radis.api.exomolapi.get_exomol_full_isotope_name` to
    get the isopologue (sorted by terrestrial abundance) ::

        from radis.api.exomolapi import get_exomol_database_list, get_exomol_full_isotope_name
        databases, recommended = get_exomol_database_list("CH4", get_exomol_full_isotope_name("CH4", 1))
        >>> ['xsec-YT10to10', 'YT10to10', 'YT34to10'], 'YT34to10'


    .. minigallery:: radis.api.exomolapi.get_exomol_database_list


    See Also
    --------
    :py:func:`~radis.api.exomolapi.get_exomol_full_isotope_name`
    """

    if isotope_full_name is None:
        raise ValueError(
            f"Give isotope name. List of known isotopes for {molecule} : {get_list_of_known_isotopes(molecule)}"
        )

    url = f"https://exomol.com/data/molecules/{molecule}/{isotope_full_name}"
    try:
        response = urlopen(url).read()
    except HTTPError as err:
        if isotope_full_name not in get_list_of_known_isotopes(molecule):
            extra = f". Isotope name {isotope_full_name} is not in list of known isotopes : {get_list_of_known_isotopes(molecule)}"
        else:
            extra = ""
        raise ValueError(f"HTTPError opening url={url}" + extra) from err

    soup = BeautifulSoup(
        response, features="lxml"
    )  # make soup that is parse-able by bs

    # Recommended database
    rows = soup.find_all(
        "a", {"class": "list-group-item link-list-group-item recommended"}
    )
    databases_recommended = [r.get_attribute_list("title")[0] for r in rows]
    databases_recommended = list(np.unique(databases_recommended))

    # All others
    rows = soup.find_all("a", {"class": "list-group-item link-list-group-item"})
    databases = [r.get_attribute_list("title")[0] for r in rows]

    if len(databases_recommended) > 1:
        # Known exceptions :
        if (
            isotope_full_name == "28Si-16O"
            and databases_recommended[0] == "xsec-SiOUVenIR"
        ):
            # this is a cross-section dataset, shouldn't be used. Reverse and use the other one:
            databases_recommended = databases_recommended[::-1]
        else:
            print(
                f"Multiple recommended databases found for {molecule} in ExoMol : {databases_recommended}. This is unexpected. Using the first"
            )

    databases = databases + databases_recommended

    if len(databases_recommended) > 0:
        recommended_database = databases_recommended[0]
    else:
        recommended_database = False

    return list(np.unique(databases)), recommended_database


# def fetch_exomol_molecule_list():
#     """Parse ExoMol website and return list of available databases, and recommended database

#     Parameters
#     ----------
#     molecule: str
#     isotope_full_name: str
#         isotope full name (ex. ``12C-1H4`` for CH4,1). Get it from
#         :py:func:`radis.api.exomolapi.get_exomol_full_isotope_name`

#     Returns
#     -------

#     Examples
#     --------
#     Get CH4 from ExoMol :
#     ::
#         databases, recommended = get_exomol_database_list("CH4", "12C-1H4")
#         >>> ['xsec-YT10to10', 'YT10to10', 'YT34to10'], 'YT34to10'

#     Or combine with :py:func:`~radis.api.exomolapi.get_exomol_full_isotope_name` to
#     get the isopologue (sorted by terrestrial abundance) ::

#         from radis.api.exomolapi import get_exomol_database_list, get_exomol_full_isotope_name
#         databases, recommended = get_exomol_database_list("CH4", get_exomol_full_isotope_name("CH4", 1))
#         >>> ['xsec-YT10to10', 'YT10to10', 'YT34to10'], 'YT34to10'

#     .. minigallery:: radis.api.exomolapi.get_exomol_database_list

#     See Also
#     --------
#     :py:func:`~radis.api.exomolapi.get_exomol_full_isotope_name`
#     """

# url = f"https://exomol.com/data/molecules/"
# try:
#     response = urlopen(url).read()
# except HTTPError as err:
#     raise ValueError(f"HTTPError opening url={url}") from err

# soup = BeautifulSoup(
#     response, features="lxml"
# )  # make soup that is parse-able by bs

# # Recommended database
# rows = soup.find_all(
#     "a", {"class": "list-group-item link-list-group-item recommended"}
# )
# databases_recommended = [r.get_attribute_list("title")[0] for r in rows]

# # All others
# rows = soup.find_all("a", {"class": "list-group-item link-list-group-item"})
# databases = [r.get_attribute_list("title")[0] for r in rows]

# if len(databases_recommended) > 1:
#     # Known exceptions :
#     if (
#         isotope_full_name == "28Si-16O"
#         and databases_recommended[0] == "xsec-SiOUVenIR"
#     ):
#         # this is a cross-section dataset, shouldn't be used. Reverse and use the other one:
#         databases_recommended = databases_recommended[::-1]
#     else:
#         print(
#             f"Multiple recommended databases found for {molecule} in ExoMol : {databases_recommended}. This is unexpected. Using the first"
#         )

# databases = databases + databases_recommended

# return databases, databases_recommended[0]


class MdbExomol(DatabaseManager):
    """molecular database of ExoMol

    MdbExomol is a class for ExoMol.

    Parameters
    ----------
    path: str
        path for Exomol data directory/tag. For instance, "/home/CO/12C-16O/Li2015"
    nurange: array
        wavenumber range list (cm-1) or wavenumber array
    margin: float
        margin for nurange (cm-1)
    crit: float
        line strength lower limit for extraction
    bkgdatm: str
        background atmosphere for broadening. e.g. H2, He,
    broadf: bool
        if False, the default broadening parameters in .def file is used

    Other Parameters
    ----------------
    engine : str
        which memory mapping engine to use : 'vaex', 'pytables' (HDF5), 'feather'
    skip_optional_data : bool
        If False, fetch all fields which are marked as available in the ExoMol definition
        file. If True, load only the first 4 columns of the states file
        ("i", "E", "g", "J"). The structure of the columns above 5 depend on the
        the definitions file (*.def) and the Exomol version.
        If ``skip_optional_data=False``, two errors may occur:

            - a field is marked as present/absent in the *.def field but is
              absent/present in the *.states file (ie both files are inconsistent).
            - in the updated version of Exomol, new fields have been added in the
              states file of some species. But it has not been done for all species,
              so both structures exist. For instance, the states file of
              https://exomol.com/data/molecules/HCl/1H-35Cl/HITRAN-HCl/ follows the
              structure described in [1]_, unlike the states file of
              https://exomol.com/data/molecules/NO/14N-16O/XABC/ which follows the
              structure described in [2]_.

    Notes
    -----

    The trans/states files can be very large. For the first time to read it,
    we convert it to the feather or hdf5-format. After the second-time,
    we use the feather/hdf5 format instead.

    Examples
    --------
    ::

        # Init database, download files if needed.
        mdb = MdbExomol(
            local_path,
            molecule=molecule,
            name=databank_name,
            local_databases=local_databases,
            # nurange=[load_wavenum_min, load_wavenum_max],
            engine="vaex",
        )

        # Get cache files to load :
        mgr = mdb.get_datafile_manager()
        local_files = [mgr.cache_file(f) for f in mdb.trans_file]

        # Load files
        df = mdb.load(
            local_files,
            columns=columns_exomol,
            lower_bound=([('nu_lines', load_wavenum_min)] if load_wavenum_min else []) + ([("Sij0", mdb.crit)] if not np.isneginf(mdb.crit) else []),
            upper_bound=([('nu_lines', load_wavenum_max)] if load_wavenum_max else []),
            output="jax", # or "pytables", "vaex"
        )

    .. minigallery:: radis.fetch_exomol


    DataFrame columns
    -----------------

    nu_lines (nd array): line center (cm-1)
    Sij0 (nd array): line strength at T=Tref (cm)
    dev_nu_lines (np array): line center in device (cm-1)
    logsij0 (np array): log line strength at T=Tref
    A (np array): Einstein A coefficient
    elower (np array): the lower state energy (cm-1)
    gpp (np array): statistical weight
    jlower (np array): J_lower
    jupper (np array): J_upper
    n_Tref (np array): temperature exponent
    alpha_ref (np array): alpha_ref (gamma0)
    n_Tref_def: default temperature exponent in .def file, used for jlower not given in .broad
    alpha_ref_def: default alpha_ref (gamma0) in .def file, used for jlower not given in .broad

    References
    ----------

    .. [1] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Barton, E. J., Chubb, K. L., Coles, P. A., … Zak, E. (2016). The ExoMol database: molecular line lists for exoplanet and other hot atmospheres. https://doi.org/10.1016/j.jms.2016.05.002
    .. [2] Tennyson, J., Yurchenko, S. N., Al-Refaie, A. F., Clark, V. H. J., Chubb, K. L., Conway, E. K., … Yurchenko, O. P. (2020). The 2020 release of the ExoMol database: Molecular line lists for exoplanet and other hot atmospheres. Journal of Quantitative Spectroscopy and Radiative Transfer, 255, 107228. https://doi.org/10.1016/j.jqsrt.2020.107228

    See also
    --------

    MdbExomol is compatible with Exojax :py:class:`exojax.spec.api.MdbExomol`

    """

    # TODO : inherit from DatabaseManager or similar

    # @dev: In exojax this class is defined in exojax/spec/moldb.py
    # see https://github.com/HajimeKawahara/exojax/blob/develop/src/exojax/spec/moldb.py
    # It uses Jax arrays (jnp). Here RADIS uses Numpy arrays.

    def __init__(
        self,
        path,
        molecule,
        database=None,
        local_databases=None,
        name="EXOMOL-{molecule}",
        nurange=[0.0, np.inf],
        margin=0.0,
        crit=-np.inf,
        bkgdatm="Air",  # TODO: use Air whenever possible (consistent with HITRAN/HITEMP). This is not a parameter for the moment.
        broadf=True,
        engine="vaex",
        verbose=True,
        cache=True,
        skip_optional_data=True,
    ):
        super().__init__(
            name,
            molecule,
            local_databases,
            engine,
            verbose=verbose,
        )
        assert cache  # cache only used for cache='regen' or cache='force' modes, cache=False is not expected

        if engine == "default":
            from radis import config

            engine = config["MEMORY_MAPPING_ENGINE"]
            if engine == "auto":
                engine = get_auto_MEMORY_MAPPING_ENGINE()

        self.path = pathlib.Path(path)
        if local_databases is not None:
            self.path = pathlib.Path(local_databases).expanduser() / self.path

        t0 = self.path.parents[0].stem
        molec = t0 + "__" + str(self.path.stem)
        self.bkgdatm = bkgdatm  # TODO: for the moment, only air is possible in RADIS, although He and H are usually available in ExoMol
        molecbroad = t0 + "__" + self.bkgdatm

        self.crit = crit
        self.margin = margin
        self.nurange = [np.min(nurange), np.max(nurange)]
        self.wmin, self.wmax = np.min(nurange), np.max(nurange)
        self.broadf = broadf
        # Where exomol files are
        self.states_file = self.path / pathlib.Path(molec + ".states.bz2")
        self.pf_file = self.path / pathlib.Path(molec + ".pf")
        self.def_file = self.path / pathlib.Path(molec + ".def")
        self.broad_file = self.path / pathlib.Path(molecbroad + ".broad")

        mgr = self.get_datafile_manager()
        if not self.def_file.exists():
            self.download(molec, extension=[".def"])
        if not self.pf_file.exists():
            self.download(molec, extension=[".pf"])
        if (
            not self.states_file.exists()
            and not mgr.cache_file(self.states_file).exists()
        ):
            self.download(molec, extension=[".states.bz2"])
        if (not self.broad_file.exists()) and self.broadf:
            self.download(molec, extension=[".broad"])

        # Add molecule name
        tag = molec.split("__")
        self.isotope_fullname = tag[0]
        self.molecule = exact_molname_exomol_to_simple_molname(tag[0])
        # self.isotope = 1  # Placeholder. TODO : implement parsing of other isotopes.

        # load def
        dic_def = read_def(self.def_file)  # approx. 3 ms
        self.n_Texp_def = dic_def["n_Texp"]
        self.alpha_ref_def = dic_def["alpha_ref"]
        self.molmass = dic_def["molmass"]

        #  default n_Texp value if not given
        if self.n_Texp_def is None:
            self.n_Texp_def = 0.5
            warnings.warn(
                Warning(
                    f"""
                    No default broadening exponent in def file. Assigned n = {self.n_Texp_def}
                    """
                )
            )
        #  default alpha_ref value if not given
        if self.alpha_ref_def is None:
            self.alpha_ref_def = 0.07
            warnings.warn(
                Warning(
                    f"""
                    No default broadening in def file. Assigned alpha_ref = {self.alpha_ref_def}
                    """
                )
            )

        # load states
        if cache == "regen" and mgr.cache_file(self.states_file).exists():
            if self.verbose:
                print("Removing existing file ", mgr.cache_file(self.states_file))
            os.remove(mgr.cache_file(self.states_file))
        if mgr.cache_file(self.states_file).exists():
            states = mgr.read(mgr.cache_file(self.states_file))
        else:
            if cache == "force":
                raise ValueError(
                    f"Cache file {str(mgr.cache_file(self.states_file))} does not exist"
                )
            print(
                f"Note: Caching states data to the {engine} format. After the second time, it will become much faster."
            )
            states = read_states(
                self.states_file,
                dic_def,
                engine="vaex" if engine == "vaex" else "csv",
                skip_optional_data=skip_optional_data,
            )
            mgr.write(mgr.cache_file(self.states_file), states)

        # load pf
        pf = read_pf(self.pf_file)
        self.gQT = pf["QT"].to_numpy()  # grid QT
        self.T_gQT = pf["T"].to_numpy()  # T forgrid QT

        # trans file(s)
        # Compute linestrengths or retrieve them from cache
        self.Tref = 296.0
        self.QTref = np.array(self.QT_interp(self.Tref))

        # Download files

        # Generate list of files
        # ---------------------
        # Case1 : transitions are stored in a single file
        if dic_def["numinf"] is None:
            self.trans_file = [self.path / pathlib.Path(molec + ".trans.bz2")]
            self.num_tag = [None]

        # Case2 : Transitions are stored in multiple files:
        else:  # dic_def["numinf"] is not None
            # dic_def["numinf"] contains the limit points ``[w(0), w(1), ..., w(n)]``
            # (n+1 elements) defining the spectral ranges appearing in the list
            # dic_def["numtag"] which looks like ``["w(0)-w(1)", "w(1)-w(2)", ..., w(n-1)-w(n)]``
            # (n elements)
            # imin :
            # index i of the "w(i)-w(i+1)" element in dic_def["numtag"] such
            # that nurange[0]<=w(i)
            imin = (
                np.searchsorted(dic_def["numinf"], nurange[0] - margin, side="right")
                - 1
            )
            # imax :
            # index i of the "w(i)-w(i+1)" element in dic_def["numtag"] such
            # that w(i+1)<=nurange[1]
            imax = (
                np.searchsorted(dic_def["numinf"], nurange[1] + margin, side="right")
                - 1
            )
            # not to exceed index out of the range
            imax = np.min([imax, len(dic_def["numinf"]) - 2])
            self.trans_file = []
            self.num_tag = []

            for k, i in enumerate(range(imin, imax + 1)):
                trans_file = self.path / pathlib.Path(
                    molec + "__" + dic_def["numtag"][i] + ".trans.bz2"
                )
                self.trans_file.append(trans_file)
                self.num_tag.append(dic_def["numtag"][i])

        # some verbose
        if self.verbose:
            print("Molecule: ", molecule)
            print("Isotopologue: ", self.isotope_fullname)
            print("Background atmosphere: ", self.bkgdatm)
            print("ExoMol database: ", database)
            print("Local folder: ", self.path)
            print("Transition files: ")

        # Look-up missing parameters and write file
        # -----------------------------------------
        for trans_file, num_tag in zip(self.trans_file, self.num_tag):
            if self.verbose:
                print(
                    "\t => File {}".format(
                        os.path.splitext(os.path.basename((trans_file)))[0]
                    )
                )

            if cache == "regen" and mgr.cache_file(trans_file).exists():
                if self.verbose:
                    print("\t\t => `regen = True`. Removing the file")
                os.remove(mgr.cache_file(trans_file))

            if not mgr.cache_file(trans_file).exists():
                if cache == "force":
                    raise ValueError(
                        f"Cache file {str(mgr.cache_file(trans_file))} does not exist"
                    )

                if not trans_file.exists():
                    self.download(molec, extension=[".trans.bz2"], numtag=num_tag)
                if self.verbose:
                    print(
                        f"\t\t => Caching the *.trans.bz2 file to the {engine} (*.h5) format. After the second time, it will become much faster."
                    )
                    print(f"\t\t => You can deleted the 'trans.bz2' file by hand.")
                trans = read_trans(
                    trans_file, engine="vaex" if engine == "vaex" else "csv"
                )
                # TODO: add option to delete file at the end

                # Complete transition data with lookup on upper & lower state :
                # In particular, compute gup and elower

                trans = pickup_gE(
                    states,
                    trans,
                    dic_def,
                    skip_optional_data=skip_optional_data,
                    engine=engine,
                )

                ##Recompute Line strength:
                from radis.lbl.base import (  # TODO: move elsewhere
                    linestrength_from_Einstein,
                )

                self.Sij0 = linestrength_from_Einstein(
                    A=trans["A"],
                    gu=trans["gup"],
                    El=trans["elower"],
                    Ia=1,  #  Sij0 is a linestrength calculated without taking into account isotopic abundance (unlike line intensity parameter of HITRAN. In RADIS this is corrected for in fetch_exomol()  )
                    nu=trans["nu_lines"],
                    Q=self.QTref,
                    T=self.Tref,
                )

                trans["Sij0"] = self.Sij0

                mgr.write(mgr.cache_file(trans_file), trans)

    def set_broadening_coef(
        self, df, alpha_ref_def=None, n_Texp_def=None, output=None, add_columns=True
    ):
        """setting broadening parameters

        Parameters
        ----------
        df: Data Frame
        alpha_ref: set default alpha_ref and apply it. None=use self.alpha_ref_def
        n_Texp_def: set default n_Texp and apply it. None=use self.n_Texp_def
        add_columns: adds alpha_ref and n_Texp columns to df

        Returns
        -------
        None. Store values in Data Frame.

        """

        if alpha_ref_def:
            self.alpha_ref_def = alpha_ref_def
        if n_Texp_def:
            self.n_Texp_def = n_Texp_def

        if self.broadf and os.path.exists(self.broad_file):
            bdat = read_broad(self.broad_file)
            if self.verbose > 1:
                print(
                    "The file `{}` is used.".format(os.path.basename(self.broad_file))
                )

            codelv = check_code_level(bdat)
            if self.verbose:
                print("Broadening code level:", codelv)

            if codelv == "a0":
                j2alpha_ref, j2n_Texp = make_j2b(
                    bdat,
                    alpha_ref_default=self.alpha_ref_def,
                    n_Texp_default=self.n_Texp_def,
                    jlower_max=df["jlower"].max(),
                )
                self.alpha_ref = j2alpha_ref[df["jlower"].values]
                self.n_Texp = j2n_Texp[df["jlower"].values]
            elif codelv == "a1":
                j2alpha_ref, j2n_Texp = make_j2b(
                    bdat,
                    alpha_ref_default=self.alpha_ref_def,
                    n_Texp_default=self.n_Texp_def,
                    jlower_max=df["jlower"].max(),
                )
                jj2alpha_ref, jj2n_Texp = make_jj2b(
                    bdat,
                    j2alpha_ref_def=j2alpha_ref,
                    j2n_Texp_def=j2n_Texp,
                    jupper_max=df["jupper"].max(),
                )
                self.alpha_ref = np.array(
                    jj2alpha_ref[df["jlower"].values, df["jupper"].values]
                )
                self.n_Texp = np.array(
                    jj2n_Texp[df["jlower"].values, df["jupper"].values]
                )
            else:
                warnings.warn(
                    f"The broadening file contains this broadening code: {codelv}."
                    + " This broadening code is NOT implemented yet.\n"
                    + "Using default parameters instead."
                )
                self.alpha_ref = np.array(self.alpha_ref_def * np.ones(len(df)))
                self.n_Texp = np.array(self.n_Texp_def * np.ones(len(df)))
        else:
            if not os.path.exists(self.broad_file):
                warnings.warn(
                    "Could not load `{}`. The default broadening parameters are used.\n".format(
                        os.path.basename(self.broad_file)
                    )
                )
            print("The default broadening parameters are used.")

            self.alpha_ref = np.array(self.alpha_ref_def * np.ones(len(df)))
            self.n_Texp = np.array(self.n_Texp_def * np.ones(len(df)))

        if add_columns:
            # Add values
            self.add_column(df, "alpha_ref", self.alpha_ref)
            self.add_column(df, "n_Texp", self.n_Texp)

    def QT_interp(self, T):
        """interpolated partition function

        Parameters
        ----------
        T: temperature

        Returns
        -------
        Q(T) interpolated in jnp.array

        """
        return np.interp(T, self.T_gQT, self.gQT)

    def qr_interp(self, T):
        """interpolated partition function ratio

        Parameters
        ----------
        T: temperature

        Returns
        -------
        qr(T)=Q(T)/Q(Tref) interpolated in jnp.array

        """
        return self.QT_interp(T) / self.QT_interp(self.Tref)

    def download(self, molec, extension, numtag=None):
        """Downloading Exomol files

        Parameters
        ----------
        molec: like "12C-16O__Li2015"
        extension: extension list e.g. [".pf",".def",".trans.bz2",".states.bz2",".broad"]
        numtag: number tag of transition file if exists. e.g. "11100-11200"

        Notes
        -----
        The download URL is written in exojax.utils.url.

        """
        import os
        import urllib.request

        tag = molec.split("__")
        molname_simple = exact_molname_exomol_to_simple_molname(tag[0])
        # TODO: add progress bar
        for ext in extension:
            if ext == ".trans.bz2" and numtag is not None:
                ext = "__" + numtag + ext

            if ext == ".broad":
                pfname_arr = [
                    tag[0] + "__H2" + ext,
                    tag[0] + "__He" + ext,
                    tag[0] + "__air" + ext,
                ]
                url = EXOMOL_URL + molname_simple + "/" + tag[0] + "/"
            else:
                pfname_arr = [molec + ext]
                url = EXOMOL_URL + molname_simple + "/" + tag[0] + "/" + tag[1] + "/"

            for pfname in pfname_arr:
                pfpath = url + pfname
                os.makedirs(str(self.path), exist_ok=True)
                if self.verbose:
                    print(
                        "\t\t => Downloading from {}".format(pfpath)
                    )  # modify indent accordingly print in __init__
                try:
                    urllib.request.urlretrieve(pfpath, str(self.path / pfname))
                except HTTPError:
                    print(f"Error: Couldn't download {ext} file at {pfpath} and save.")

    def to_partition_function_tabulator(self):
        """Generate a :py:class:`~radis.levels.partfunc.PartFuncExoMol` object"""
        from radis.levels.partfunc import PartFuncExoMol

        return PartFuncExoMol(self.isotope_fullname, self.T_gQT, self.gQT)


if __name__ == "__main__":
    print(exact_molname_exomol_to_simple_molname("12C-1H4"))
    print(exact_molname_exomol_to_simple_molname("23Na-16O-1H"))
    print(exact_molname_exomol_to_simple_molname("HeH_p"))
    print(exact_molname_exomol_to_simple_molname("trans-31P2-1H-2H"))  # not working

    # TODO : move to unitary tests of exomol_utils
    assert exact_molname_exomol_to_simple_molname("12C-1H4") == "CH4"
    assert exact_molname_exomol_to_simple_molname("23Na-16O-1H") == "NaOH"
    assert exact_molname_exomol_to_simple_molname("HeH_p") == "HeH_p"
    assert (
        exact_molname_exomol_to_simple_molname("trans-31P2-1H-2H") == "trans-31P2-1H-2H"
    )  # convert not working

    assert exact_molname_exomol_to_simple_molname("12C-16O2") == "CO2"
    assert exact_molname_exomol_to_simple_molname("13C-16O2") == "CO2"

    # mdb=MdbExomol("/home/kawahara/exojax/data/CO/12C-16O/Li2015/")
    # mdb=MdbExomol("/home/kawahara/exojax/data/CH4/12C-1H4/YT34to10/",nurange=[6050.0,6150.0])

    # %% Test  by overriding Spectrumfactory's DataFrame df0
    # mdb = MdbExomol(".database/H2O/1H2-16O/POKAZATEL", [4310.0, 4320.0], crit=1.0e-45)
    # df = mdb.to_df()
    # from radis import SpectrumFactory
    # sf = SpectrumFactory(
    #     4310,
    #     4320,
    #     molecule="H2O",
    #     isotope="1",
    # )
    # sf.fetch_databank(
    #     "hitran"
    # )  # placeholder. Load lines (will be replaced), load partition function.
    # sf.df0 = df  # override.
    # s = sf.eq_spectrum(500, name="ExoMol")
    # # sf.fetch_databank('hitran')  # placeholder. Load lines (will be replaced), load partition function.
    # # s_hit = sf.eq_spectrum(500, name='HITRAN')

    # %% Test by direct calculation
    # import pytest

    # print("Testing factory:", pytest.main(["../test/io/test_exomol.py"]))

    # %% RADIS-like Example
    # uses fetch_exomol() internally

    from radis import calc_spectrum

    s = calc_spectrum(
        wavelength_min=1.630e4,
        wavelength_max=1.6305e4,
        molecule="CH4",
        isotope="1",
        pressure=1.01325,  # bar
        Tgas=1000,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        databank=(
            "exomol",
            "YT10to10",
        ),  # Simply use 'exomol' for the recommended database
    )
    # s.apply_slit(1, "cm-1")  # simulate an experimental slit
    s.plot("xsection")

    # %% Exojax like Example

    class mdbExoMol:

        # hardcode attribute names, to prevent typos and the declaration of unwanted parameters
        __slots__ = [
            "Sij0",
            "logsij0",
            "nu_lines",
            "A",
            "elower",
            "eupper",
            "gupper",
            "jlower",
            "jupper",
        ]

        def __init__(
            self,
            molecule,
            path,
            nurange=[-np.inf, np.inf],
            local_databases="~/exojax",
        ):
            """
            Parameters
            ----------
            molecule: molecule name
            path : local path, mirror of ExoMol path
            nurange : TYPE, optional
                DESCRIPTION. The default is [-np.inf, np.inf].
            crit : TYPE, optional
                DESCRIPTION. The default is -np.inf.

            Returns
            -------
            DataFrame

            Examples
            --------
            ::

                mdbCH4 = mdbExoMol("CH4", '.database/CH4/12C-1H4/YT10to10/', nus, crit=1.e-30)
                print(len(mdbCH4.nu_lines), "lines")
                mdbCH4.elower

            Available columns::

                [
                    "Sij0",
                    "logsij0",
                    "nu_lines",
                    "A",
                    "elower",
                    "eupper",
                    "gupper",
                    "jlower",
                    "jupper",
                ]

            """

            wavenum_min, wavenum_max = np.min(nurange), np.max(nurange)
            if wavenum_min == -np.inf:
                wavenum_min = None
            if wavenum_max == np.inf:
                wavenum_max = None

            # Set-up database, download files and set-up cache files if needed
            mdb = MdbExomol(
                path,
                molecule=molecule,
                local_databases=local_databases,
                nurange=[wavenum_min, wavenum_max],
            )

            # Get cache files to load :
            mgr = mdb.get_datafile_manager()
            local_files = [mgr.cache_file(f) for f in mdb.trans_file]

            # Load them: #temperature dependent criterion
            jdict = mdb.load(
                local_files,
                columns=[k for k in self.__slots__ if k not in ["logsij0"]],
                lower_bound=([("nu_lines", wavenum_min)] if wavenum_min else [])
                + ([("Sij0", mdb.crit)] if not np.isneginf(mdb.crit) else []),
                upper_bound=([("nu_lines", wavenum_max)] if wavenum_max else []),
                output="jax",
            )

            # set attributes, accessible as e.g:  mdb.nu_lines
            for k in jdict.keys():
                setattr(self, k, jdict[k])

    nus = np.linspace(1e7 / 1.630e4, 1e7 / 1.6305e4)

    # Download new ExoMol repo (in ~/exomol)
    mdbCH4 = mdbExoMol(
        "CH4",
        ".database/CH4/12C-1H4/YT10to10/",
        nus,
        crit=1.0e-30,
        local_databases=".",  # use local folder
    )

    print(len(mdbCH4.nu_lines), "lines")
    mdbCH4.elower

    # Or use RADIS's folder  (# by default ~/.radisdb/exomol)
    import radis

    mdbCH4_2 = mdbExoMol(
        "CH4",
        "CH4/12C-1H4/YT10to10/",
        nus,
        crit=1.0e-30,
        local_databases=pathlib.Path(radis.config["DEFAULT_DOWNLOAD_PATH"]) / "exomol",
    )
    # ... ready to run Jax calculations

    # %%

    # """ExoMol lines can be downloaded and accessed separately using
    # :py:func:`~radis.io.exomol.fetch_exomol`
    # """

    # # See line data:
    # from radis.io.exomol import fetch_exomol

    # df = fetch_exomol("SiO", database="EBJT", isotope="1", load_wavenum_max=5000)
    # print(df)

    # #%%
    # # See the list of recommended databases for the 1st isotope of SiO :
    # from radis.api.exomolapi import get_exomol_database_list, get_exomol_full_isotope_name

    # databases, recommended = get_exomol_database_list(
    #     "SiO", get_exomol_full_isotope_name("SiO", 1)
    # )
    # print("Databases for SiO: ", databases)
    # print("Database recommended by ExoMol: ", recommended)



# -*- coding: utf-8 -*-
"""Variants of default Partition function tabulators and calculators, based on
CDSD-4000 format, specifically for CO2.

Routine Listing
---------------

Partition functions:

- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDcalc`
- :class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDtab`

Which inherit from:

- :class:`~radis.levels.partfunc.RovibParFuncCalculator`
- :class:`~radis.levels.partfunc.RovibParFuncTabulator`


-------------------------------------------------------------------------------
"""


import time
from os.path import exists, getmtime
from warnings import warn

import pandas as pd
from scipy.interpolate import splev, splrep

import radis
from radis.api.cache_files import filter_metadata, load_h5_cache_file, save_to_hdf
from radis.db.molecules import ElectronicState
from radis.lbl.labels import (
    vib_lvl_name_cdsd_p,
    vib_lvl_name_cdsd_pc,
    vib_lvl_name_cdsd_pcJN,
    vib_lvl_name_cdsd_pcN,
)
from radis.levels.partfunc import RovibParFuncCalculator, RovibParFuncTabulator
from radis.misc.warning import OutOfBoundError
from radis.test.utils import getTestFile

# %% Variants of Tabulated partition functions (interpolate)


class PartFuncCO2_CDSDtab(RovibParFuncTabulator):
    """Return partition function of CO2 using a spline interpolation of
    tabulated values used in [CDSD-4000]_

    Parameters
    ----------

    T: temperature (K)
        gas temperature (equilibrium)

    Notes
    -----

    Partition function calculated in CDSD by direct summation (Jmax=300)

    The CDSD ``partition_functions.txt`` can be downloaded from the
    [CDSD-4000]_ FTP : ftp://ftp.iao.ru/pub/CDSD-4000/

    See Also
    --------

    :py:class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDcalc`
    """

    def __init__(self, isotope, database):
        r"""Get partition function for one isotope only.

        (note that this forces to reload the file once per isotope, but
        at least we have a clean layout with one object per isotope)
        """

        Itable = {1: "12C16O2", 2: "13C16O2", 3: "16O12C18O", 4: "16O12C17O"}

        # Check input
        if not isotope in Itable:
            raise KeyError(
                "Isotope {0} not defined in CDSD tabulated ".format(isotope)
                + "partition functions. Only the following are: {0}".format(
                    list(Itable.keys())
                )
            )

        # Read partition function tabulated data
        parsum = pd.read_csv(database, comment="#", sep="\s+")
        if not "T(K)" in list(parsum.keys()):
            raise KeyError("Missing columns ({0}) in {1}".format("T(K)", database))

        # Define spline interpolation
        isoname = Itable[isotope]
        tck = splrep(parsum["T(K)"], parsum[isoname])

        self.molecule = 2  # 'CO2'
        self.iso = isotope

        self.tck = tck  # dictionary
        self.Tmin = parsum["T(K)"].min()
        self.Tmax = parsum["T(K)"].max()

    def _inrange(self, T):
        r"""Allow for 5% extrapolation (ex: 296K / 300K) )"""
        return (self.Tmin * 0.95 <= T) and (self.Tmax * 1.05 >= T)

    def _at(self, T):
        r"""Get partition function at temperature T.

        Called by :meth:`radis.levels.partfunc.RovibParFuncTabulator.at`
        """
        if not self._inrange(T):
            raise OutOfBoundError(
                "Temperature: {0} is out of bounds {1}-{2}".format(
                    T, self.Tmin, self.Tmax
                )
            )

        return splev(T, self.tck)


# %% Calculated partition functions (either from energy levels, or ab initio)


class PartFuncCO2_CDSDcalc(RovibParFuncCalculator):
    """Calculate Partition Function from energy levels (and maybe export a
    tabulated database).

    warning::
        ZPE (zero point energy) must be the same in the Line Database and the
        Energy levels database. See the
        :ref:`Configuration file <label_lbl_config_file>`.

    Parameters
    ----------
    energy_levels: filename
        path to energy levels (to calculate Partition Function) for ``isotope``
    isotope: int
        which isotope we're dealing with. Default ``1``. In the current implementation
        only isotope 1 and 2 are defined.
    levelsfmt: ``'cdsd-p'``, ``'cdsd-pc'``, ``'cdsd-pcN'``, ``'cdsd-hamil'``, or ``None``
        the format of the Energy Database, and in particular how ``Evib`` and ``Erot``
        have been calculated. A vibrational level in the CDSD (p,c,J,N) nomenclature
        can be defined for levels that share a same (p), (p,c) or (p,c,N), where
        ``p`` is the polyad number, ``c`` is the Wang symmetry number, and ``N``
        is the ranking index of a (p,c,J) group. Default ``'cdsd-pc'``.

        If ``None``, dont label the levels. Wont be able to use the EnergyDatabase to fetch
        vibrational energies for lines, however it can still be used to
        calculate Partition functions independently from a Spectrum calculation

    Other Parameters
    ----------------
    use_cached: ``True``, ``False``, or ``'regen'``, ``'force'``
        if ``True``, use (and generate if doesnt exist) a ``.h5`` file.
        If ``'regen'``, regenerate cache file. If ``'force'``, raise an error
        if file doesnt exist. Default ``True``
    mode: 'full summation', 'tabulation'
        calculation mode. ``'tabulation'`` is much faster but not all possible
        distributions are implemented. Default ``'full-summation'``
    use_json: boolean
        deprecated. Better use h5 now.

    Notes
    -----

    Database format:

    Tashkun database updated with ranking number (n) & total rank (N) of
    block, Evib and Erot (cm-1)  and jref

    For nonequilibrium, different strategies exist so as how to assign rotational and vibrational
    energies in a CDSD database. See the E. Pannier "Limits on CO2 Nonequilibrium
    model" article for a discussion on that.

    Example of table format::

        # calculated rovibrational energy levels of 12C16O2
        # =================================================
        # S.Tashkun, Zuev Institute of Atmospheric Optics, Tomsk, Russia
        # date: 17.03.2017
        #
        # zero point energy ZPE (cm-1) =  2531.828
        #
        # p = 2v1 + v2 + 3v3 - polyad number
        # j - rotational quantum number
        # c - Wang symmetry (1-'e'; 2-'f')
        # N - ranking number of energy levels of (p,j,c) blocks
        # n - total rank of (p,j,c) blocks
        #
        # Calculation limitations:
        # pmax = 40
        # jmax = 300
        # Ecut = 44600 cm-1
        # ---------------
        p   c   j   N   n   E   Evib    Erot    jref
        0   1   0   1   1   0.000   0.000   0.000   0
        0   1   2   1   1   2.341   0.000   2.341   0
        0   1   4   1   1   7.804   0.000   7.804   0
        0   1   6   1   1   16.389  0.000   16.389  0
        0   1   8   1   1   28.095  0.000   28.095  0

    See an example in `test/files/co2_cdsd_hamiltonian_fragment.levels <https://github.com/radis/radis/blob/develop/radis/test/files/cdsd_hitemp_09_fragment.txt>`

    See Also
    --------

    :py:class:`~radis.levels.partfunc_cdsd.PartFuncCO2_CDSDtab`
    """

    def __init__(
        self,
        energy_levels,
        isotope,
        levelsfmt,  # ='cdsd-pc',
        use_cached=True,
        use_json=None,  # TODO: Deprecated, remove
        verbose=True,
        mode="full summation",
    ):

        # %% Init

        # Initialize PartitionFunctionCalculator for this electronic state
        ElecState = ElectronicState("CO2", isotope, "X", "1Σu+")
        super(PartFuncCO2_CDSDcalc, self).__init__(ElecState, mode=mode)

        # Check inputs ('return' is not mentioned in signature. it will just return
        # after cache name is given)
        assert use_cached in [True, False, "regen", "force", "return"]
        if isotope not in [1, 2]:
            raise ValueError(
                "CDSD Energies not defined for isotope: {0}".format(isotope)
            )
        if use_json is not None:
            warn(
                DeprecationWarning(
                    "use_json replaced with faster HDF5-based use_cached"
                )
            )
        # Get vibrational level definitions that match Energy Database (in particular
        # how Evib and Erot are calculated)
        # This is needed to be able to match the levels in the Line Database and
        # the levels in the Energy database
        if levelsfmt == "cdsd-p":
            viblvl_label = "p"
        elif levelsfmt == "cdsd-pc":
            viblvl_label = "pc"
        elif levelsfmt == "cdsd-pcN":
            viblvl_label = "pcN"
        elif levelsfmt == "cdsd-hamil":
            viblvl_label = "pcJN"
        elif levelsfmt is None:
            # dont label the levels. Wont be able to use the EnergyDatabase to fetch
            # vibrational energies for lines, however it can still be used to
            # calculate Partition functions independently from a Spectrum calculation
            viblvl_label = None
        else:
            raise ValueError(
                "Unknown Energy database format: levelsfmt = `{0}`".format(levelsfmt)
                + ". Use one of: `cdsd-p`, `cdsd-pc`, `cdsd-pcN`,`cdsd-hamil`"
            )

        # Store defaults
        self.verbose = verbose
        self.use_cached = use_cached
        self.levelsfmt = levelsfmt
        self.viblvl_label = viblvl_label
        self.last_modification = time.ctime(
            getmtime(getTestFile(r"co2_cdsd_hamiltonian_fragment.levels"))
        )
        if verbose >= 2:
            print("Last modification time: {0}".format(self.last_modification))

        # Get variables to store in metadata  (after default values have been set)
        molecule = "CO2"  # will be stored in cache file metadata
        last_modification = time.ctime(
            getmtime(getTestFile(r"co2_cdsd_hamiltonian_fragment.levels"))
        )

        _discard = [
            "self",
            "energy_levels",
            "verbose",
            "ElecState",
            "electronic_state",
            "use_json",
            "use_cached",
        ]
        # (dev) locals() automatically stores all variables: levelsfmt, viblvl_label, etc.
        metadata = filter_metadata(locals(), discard_variables=_discard)

        # %% Get levels
        # Function of use_cached value:
        # ... if True, use (and generate if doesnt exist) cache file.
        # ... if 'regen', regenerate cache file. If 'force', raise an error
        # ... if file doesnt exist.
        # If file is deprecated, regenerate it unless 'force' was used

        # Load cache file if exists

        cachefile = energy_levels + ".h5"
        self.cachefile = cachefile

        # If return, return after cachefile generated (used for tests)
        if use_cached == "return":
            return

        df = load_h5_cache_file(
            cachefile,
            use_cached,
            valid_if_metadata_is=metadata,
            relevant_if_metadata_above={},
            relevant_if_metadata_below={},
            current_version=radis.__version__,
            last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
            verbose=verbose,
        )

        if df is None:  # Read normal file
            df = pd.read_csv(energy_levels, comment="#", sep="\s+")
            df = self._add_degeneracies(df)
            df = self._add_levels(df)

        self.df = df  # Store

        if use_cached and not exists(cachefile):
            save_to_hdf(
                self.df,
                cachefile,
                metadata=metadata,
                version=radis.__version__,
                key="df",
                overwrite=True,
                verbose=verbose,
            )

    def _add_degeneracies(self, df):
        r"""Calculate and store degeneracies in database df.

        Parameters
        ----------

        df: pandas Dataframe
            energy database

        Notes
        -----

        .. warning::

            we use the same energies as CO2 626 but with gi = 2 for CO2 isotope 2 (636).
            It is done in the __init__ method
        """
        # Rotational degeneracy
        gj = 2 * df.j + 1
        # ... state dependant (forbidden rotational level are not in the database):
        gs = 1
        # ... state independent:
        gi = self.gi()
        grot = gj * gs * gi

        # Vibrational degeneracy
        gvib = 1  # CDSD is rovibrational complete

        # Store
        df["gvib"] = gvib
        df["gj"] = gj
        df["grot"] = grot

        return df

    def _add_levels(self, df):

        viblvl_label = self.viblvl_label

        if viblvl_label == "p":
            df["viblvl"] = vib_lvl_name_cdsd_p(
                df.p,
            )
        elif viblvl_label == "pc":
            df["viblvl"] = vib_lvl_name_cdsd_pc(df.p, df.c)
        elif viblvl_label == "pcN":
            df["viblvl"] = vib_lvl_name_cdsd_pcN(df.p, df.c, df.N)
        elif viblvl_label == "pcJN":
            df["viblvl"] = vib_lvl_name_cdsd_pcJN(df.p, df.c, df.j, df.N)
        elif viblvl_label is None:
            # dont label the levels. Wont be able to use the EnergyDatabase to fetch
            # vibrational energies for lines, however it can still be used to
            # calculate Partition functions independently from a Spectrum calculation
            pass
        else:
            raise ValueError("Unexpected viblvl_label value: {0}".format(viblvl_label))

        return df

    def gs(self):
        from radis.db.degeneracies import gs

        I = self.isotope
        return gs(2, I)

    def gi(self):
        from radis.db.degeneracies import gi

        I = self.isotope
        return gi(2, I)


# %% Test
if __name__ == "__main__":

    from radis.test.levels.test_partfunc import _run_testcases

    print("Testing parfunc: {0}".format(_run_testcases()))
